"""

:mod:`Waveforms` -- Known current Waveforms for airborne EM systems
==================================================================

XXX


"""

import numpy as np
from scipy.integrate import fixed_quad
from scipy.special import roots_legendre


class CurrentWaveforms:
    """Simple Class for CurrentWaveforms."""

    def __init__(self, name):
        """Add the filter name."""
        self.name = name


def skytem_HM_2015():
    """
        SkyTEM High moment (HM) current waveform
    """

    waveform = CurrentWaveforms("skytem high moment 2015")

    waveform.base_frequency = "30 Hz"
    waveform.currrent_amplitude = "122.5 A"
    waveform.current_times = np.array(
        [
            -2.06670e-02,
            -2.05770e-02,
            -2.04670e-02,
            -1.66670e-02,
            -1.64726e-02,
            -1.64720e-02,
            -1.64706e-02,
            -4.00000e-03,
            -3.91000e-03,
            -3.80000e-03,
            0.00000e00,
            1.94367e-04,
            1.95038e-04,
            1.96368e-04,
        ]
    )

    waveform.currents = np.array(
        [
            0.00000e00,
            -5.30000e-01,
            -9.73000e-01,
            -1.00000e00,
            -2.81610e-03,
            -1.44356e-03,
            0.00000e00,
            0.00000e00,
            5.30000e-01,
            9.73000e-01,
            1.00000e00,
            2.81610e-03,
            1.44356e-03,
            0.00000e00,
        ]
    )

    # For Trapezoidal Waveform
    t0, t1, t2, t3 = -4.00000e-03, -3.80000e-03, 0.0, 1.96368e-04
    waveform.times_trapezoids = np.array([t0, t1, t2, t3])

    waveform.time_shift = 220 * 1e-6
    waveform.time_gate_center = (
        np.array(
            [
                12.715,
                16.215,
                20.715,
                26.215,
                33.215,
                42.215,
                53.715,
                68.215,
                86.215,
                108.715,
                136.715,
                172.215,
                217.715,
                274.715,
                346.715,
                437.715,
                551.715,
                695.715,
                877.215,
                1105.715,
                1394.215,
                1758.215,
                2216.715,
                2794.715,
                3523.715,
                4442.715,
                5601.215,
                7061.215,
                8902.215,
                11064.715,
            ]
        )
        * 1e-6
    )

    return waveform


def skytem_LM_2015():
    """
        SkyTEM Low moment (LM) current waveform
    """

    waveform = CurrentWaveforms("skytem low moment 2015")

    waveform.base_frequency = "210 Hz"
    waveform.currrent_amplitude = "8.3 A"
    waveform.current_times = np.array(
        [
            -3.1810e-003,
            -3.1100e-003,
            -2.7860e-003,
            -2.5334e-003,
            -2.3820e-003,
            -2.3810e-003,
            -2.3798e-003,
            -2.3779e-003,
            -2.3762e-003,
            -2.3749e-003,
            -2.3733e-003,
            -2.3719e-003,
            -2.3716e-003,
            -8.0000e-004,
            -7.2902e-004,
            -4.0497e-004,
            -1.5238e-004,
            -1.0000e-006,
            0,
            1.1535e-006,
            3.0943e-006,
            4.7797e-006,
            6.1076e-006,
            7.7420e-006,
            9.0699e-006,
            9.4274e-006,
        ]
    )

    waveform.currents = np.array(
        [
            0,
            -1.0078e-001,
            -4.5234e-001,
            -7.6328e-001,
            -1.0000e000,
            -1.0000e000,
            -8.6353e-001,
            -3.4002e-001,
            -1.1033e-001,
            -4.4709e-002,
            -1.3388e-002,
            -4.4389e-003,
            0,
            0,
            1.0078e-001,
            4.5234e-001,
            7.6328e-001,
            1.0000e000,
            1.0000e000,
            8.6353e-001,
            3.4002e-001,
            1.1033e-001,
            4.4709e-002,
            1.3388e-002,
            4.4389e-003,
            0,
        ]
    )

    # For Trapezoidal Waveform
    t0, t1, t2, t3 = -8.0000e-04, -1e-6, 0.0, 9.4274e-006
    waveform.times_trapezoids = np.array([t0, t1, t2, t3])

    waveform.time_shift = 220 * 1e-6
    waveform.time_gate_center = (
        np.array(
            [
                10.215,
                12.715,
                16.215,
                20.715,
                26.215,
                33.215,
                42.215,
                53.715,
                68.215,
                86.215,
                108.715,
                136.715,
                172.215,
                217.715,
                274.715,
                346.715,
                437.715,
                551.715,
                695.715,
                877.215,
                1105.715,
                1394.215,
            ]
        )
        * 1e-6
    )

    return waveform


def piecewise_ramp(step_func, t_channels, t_currents, currents, n=20, eps=1e-10):
    """
    Computes response from piecewise linear current waveform
    with a single pulse. This basically evaluates the convolution
    between dI/dt and step-off response.

    step_func: function handle to evaluate step-off response
    t_channels: time channels when the current is on or off
    currents: input source currents
    n: Gaussian quadrature order
    """
    dt = np.diff(t_currents)
    dI = np.diff(currents)
    dIdt = dI / dt
    nt = t_currents.size
    response = np.zeros(t_channels.size, dtype=float)
    pulse_time = t_currents.max()

    for i in range(1, nt):
        t_lag = pulse_time - t_currents[i]
        time = t_lag + t_channels
        t0 = dt[i - 1]
        const = -dIdt[i - 1]
        if abs(const) > eps:
            for j, t in enumerate(time):
                # on-time
                # TODO: this is only working when we have a single ramp...
                if t < 0.0:
                    print(t + t0)
                    response[j] += fixed_quad(step_func, 0, t + t0, n=n)[0] * const
                # off-time
                else:
                    response[j] += fixed_quad(step_func, t, t + t0, n=n)[0] * const
    return response


def piecewise_ramp_fast(step_func, t_off, t_currents, currents, x, w, eps=1e-10):
    """
    Computes response from piecewise linear current waveform
    with a single pulse. This basically evaluates the convolution
    between dI/dt and step-off response.

    step_func: function handle to evaluate step-off response
    t_off: time channels when the current is off
    t_shift: t_off + T/2
    currents: input source currents
    n: Gaussian quadrature order
    """
    n = x.size
    dt = np.diff(t_currents)
    dI = np.diff(currents)
    dIdt = dI / dt
    nt = t_currents.size
    pulse_time = t_currents.max()

    # Create a bunch of memory in C and use broadcasting
    t_lag = pulse_time - t_currents
    t_lag_expand = (np.repeat(t_lag[1:, np.newaxis], t_off.size, 1)).T
    t_lag_3D = np.repeat(t_lag_expand[:, :, np.newaxis], n, 2)
    t3D = t_lag_3D + t_off[:, np.newaxis, np.newaxis]
    # Gauss-Legendre part.
    # Expand time shifts and origin to 3D with G-L points
    inds = t3D[:, :, 0] < 0.0
    # Compute dt for both on-time and off-time
    # off-time f(t, t+t0)
    # on-time f(0, t+t0)
    dt_on_off = np.tile(dt, (t_off.size, 1))
    dt_on_off[inds] = (dt + t3D[:, :, 0])[inds]
    t3D[inds, :] = 0.0

    y = dt_on_off[:, :, np.newaxis] * (0.5 * (x + 1.0)) + t3D

    # Evaluate and weight G-L values with current waveform
    f = w * step_func(np.log10(y))
    s = f.sum(axis=2) * 0.5 * dt_on_off

    response = np.sum(s * -dIdt, axis=1)

    return response


def piecewise_ramp_fast_diff(
    step_func, t_off, t_shift, t_currents, currents, x, w, eps=1e-10
):
    """
    Computes response from piecewise linear current waveform
    with a single pulse. This basically evaluates the convolution
    between dI/dt and step-off response.

    step_func: function handle to evaluate step-off response
    t_off: time channels when the current is off
    t_shift: t_off + T/2
    currents: input source currents
    n: Gaussian quadrature order
    """
    n = x.size
    dt = np.diff(t_currents)
    dI = np.diff(currents)
    dIdt = dI / dt
    nt = t_currents.size
    pulse_time = t_currents.max()

    # Create a bunch of memory in C and use broadcasting
    t_lag = pulse_time - t_currents
    t_lag_expand = (np.repeat(t_lag[1:, np.newaxis], t_off.size, 1)).T
    t_lag_3D = np.repeat(t_lag_expand[:, :, np.newaxis], n, 2)
    t3D = t_lag_3D + t_off[:, np.newaxis, np.newaxis]

    # Gauss-Legendre part.
    # Expand time shifts and origin to 3D with G-L points
    inds = t3D[:, :, 0] < 0.0
    # Compute dt for both on-time and off-time
    # off-time f(t, t+t0)
    # on-time f(0, t+t0)
    dt_on_off = np.tile(dt, (t_off.size, 1))
    dt_on_off[inds] = (dt + t3D[:, :, 0])[inds]
    t3D[inds, :] = 0.0

    y = dt_on_off[:, :, np.newaxis] * (0.5 * (x + 1.0)) + t3D
    # Evaluate and weight G-L values with current waveform
    f = w * step_func(np.log10(y))
    s = f.sum(axis=2) * 0.5 * dt_on_off

    response = np.sum(s * -dIdt, axis=1)

    t3D = t_lag_3D + t_shift[:, np.newaxis, np.newaxis]
    inds = t3D[:, :, 0] < 0.0
    # Compute dt for both on-time and off-time
    # off-time f(t, t+t0)
    # on-time f(0, t+t0)
    dt_on_off = np.tile(dt, (t_off.size, 1))
    dt_on_off[inds] = (dt + t3D[:, :, 0])[inds]
    t3D[inds, :] = 0.0

    # Gauss-Legendre part.
    # Expand time shifts and origin to 3D with G-L points
    y = dt_on_off[:, :, np.newaxis] * (0.5 * (x + 1.0)) + t3D
    # Evaluate and weight G-L values with current waveform
    f = w * step_func(np.log10(y))
    s = f.sum(axis=2) * 0.5 * dt_on_off

    response -= 0.5 * np.sum(s * -dIdt, axis=1)

    return response


def piecewise_pulse(step_func, t_off, t_currents, currents, T, n=20, n_pulse=2):
    """
    Computes response from double pulses (negative then positive)
    T: Period (e.g. 25 Hz base frequency, 0.04 s period)
    """
    if n_pulse == 1:
        response = piecewise_ramp(step_func, t_off, t_currents, currents, n=n)
    elif n_pulse == 2:
        response = (
            piecewise_ramp(step_func, t_off, t_currents, currents, n=n)
            - piecewise_ramp(step_func, t_off + T / 2.0, t_currents, currents, n=n)
            * 0.5
        )
    else:
        raise NotImplementedError("n_pulse must be either 1 or 2")
    return response


def piecewise_pulse_fast(step_func, t_off, t_currents, currents, T, n=20, n_pulse=2):
    """
    Computes response from double pulses (negative then positive)
    T: Period (e.g. 25 Hz base frequency, 0.04 s period)
    """

    # Use early out scheme for speed. Can turn assertions off with "python -O"
    assert n_pulse == 1 or n_pulse == 2, NotImplementedError(
        "n_pulse must be either 1 or 2"
    )

    # Get gauss-legendre points and weights early since n never changes inside here
    x, w = roots_legendre(n)

    if n_pulse == 1:
        response = piecewise_ramp_fast(step_func, t_off, t_currents, currents, x, w)
    elif n_pulse == 2:
        response = piecewise_ramp_fast_diff(
            step_func, t_off, t_off + 0.5 * T, t_currents, currents, x, w
        )
    return response


def butter_lowpass_filter(highcut_frequency, fs=1e6, period=0.04, order=1):
    """
    Butterworth low pass filter

    Parameters
    ----------

    highcut_frequency: float
        high-cut frequency for the low pass filter
    fs: float
        sampling rate, 1./ dt, (default = 1MHz)
    period:
        period of the signal (e.g. 25Hz base frequency, 0.04s)
    order: int
        The order of the butterworth filter

    Returns
    -------

    frequency, h: ndarray, ndarray
        Filter values (`h`) at frequencies (`frequency`) are provided.
    """

    # Nyquist frequency
    nyq = 0.5 * fs
    n_samples = period * fs
    high = highcut_frequency / nyq
    b, a = butter(order, high, btype="low")
    w, h = freqz(b, a, worN=int(n_samples))
    frequency = (fs * 0.5 / np.pi) * w

    return frequency, h


def butterworth_type_filter(frequency, highcut_frequency, order=2):
    """
    Butterworth low pass filter

    Parameters
    ----------

    highcut_frequency: float
        high-cut frequency for the low pass filter
    fs: float
        sampling rate, 1./ dt, (default = 1MHz)
    period:
        period of the signal (e.g. 25Hz base frequency, 0.04s)
    order: int
        The order of the butterworth filter

    Returns
    -------

    frequency, h: ndarray, ndarray
        Filter values (`h`) at frequencies (`frequency`) are provided.
    """

    # Nyquist frequency
    h = 1.0 / (1 + 1j * (frequency / highcut_frequency)) ** order
    highcut_frequency = 300 * 1e3
    h *= 1.0 / (1 + 1j * (frequency / highcut_frequency)) ** 1
    return h


def rotate_origin_only(xy, radians):
    """Only rotate a point around the origin (0, 0)."""
    xx = xy[:, 0] * np.cos(radians) + xy[:, 1] * np.sin(radians)
    yy = -xy[:, 0] * np.sin(radians) + xy[:, 1] * np.cos(radians)
    return np.c_[xx, yy]


def rotate_to_x_axis(xy, observation_point):
    """
    Moves each of two sequential points to origin (0,0)
    then rotate to x-axis
    """
    n_pts = xy.shape[0]
    dxy = np.empty((n_pts - 1, 2), dtype=float, order="F")
    dxy_obs = np.empty((n_pts - 1, 2), dtype=float, order="F")
    angle = np.empty(n_pts - 1, dtype=float)
    dxy = np.diff(xy, axis=0)
    print(dxy)
    dxy_obs = observation_point.repeat(n_pts - 1).reshape((2, n_pts - 1)).T
    angle = np.arctan2(dxy[:, 1], dxy[:, 0])
    # this is for self-check
    xy_rot = rotate_origin_only(dxy, angle)
    xy_obs_rot = rotate_origin_only(dxy_obs, angle)

    # For computation of Hx, Hy angle is required
    # Hx: hx cos theta + hy sin theta
    # Hy: hx sin theta - hy cos theta
    # Hz: hz
    return xy_rot, xy_obs_rot, angle


def get_geotem_wave(pulse_period, n_segment=3):
    t = 0.0
    T = pulse_period
    time_input_currents = np.r_[0.0, T / 2.0, np.linspace(3 * T / 4.0, T, n_segment)]
    input_currents = np.sin(np.pi / T * time_input_currents)
    return time_input_currents, input_currents


def get_nanotem_wave(ramp_off_time):
    time_input_currents = np.r_[0, ramp_off_time]
    input_currents = np.r_[1, 0.0]
    return time_input_currents, input_currents


def get_flight_direction_from_fiducial(fiducial, lines, easting, northing):
    lines_unique = np.unique(lines)
    n_line = lines_unique.size
    flight_direction = np.empty(fiducial.size, dtype=float)
    i_start = 0
    for i_line, line_unique in enumerate(lines_unique):
        ind_line = lines == line_unique
        x0, x1 = easting[ind_line][0], easting[ind_line][-1]
        y0, y1 = northing[ind_line][0], northing[ind_line][-1]
        dx = x1 - x0
        dy = y1 - y0
        n = ind_line.sum()
        flight_direction[i_start : i_start + n] = np.arctan2(dy, dx)
        i_start += n
    return flight_direction


def get_rx_locations_from_flight_direction(
    flight_direction, src_locations, offset=120, dz=-45
):
    dx = offset * np.cos(np.deg2rad(flight_direction))
    dy = offset * np.sin(np.deg2rad(flight_direction))
    rx_locations = np.c_[
        src_locations[:, 0] - dx, src_locations[:, 1] - dy, src_locations[:, 2] + dz
    ]
    return rx_locations



from SimPEG import Maps, Utils, Problem, Props
import numpy as np
from .Survey import BaseEM1DSurvey
from scipy.constants import mu_0
from .RTEfun_vec import rTEfunfwd, rTEfunjac
from scipy.interpolate import InterpolatedUnivariateSpline as iuSpline

from empymod import filters
from empymod.transform import dlf, get_dlf_points
from empymod.utils import check_hankel

try:
    from simpegEM1D.m_rTE_Fortran import rte_fortran
except ImportError as e:
    rte_fortran = None


class EM1D(Problem.BaseProblem):
    """
    Pseudo analytic solutions for frequency and time domain EM problems
    assumingLayered earth (1D).
    """

    surveyPair = BaseEM1DSurvey
    mapPair = Maps.IdentityMap
    chi = None
    hankel_filter = "key_101_2009"  # Default: Hankel filter
    hankel_pts_per_dec = None  # Default: Standard DLF
    verbose = False
    fix_Jmatrix = False
    _Jmatrix_sigma = None
    _Jmatrix_height = None
    _pred = None

    sigma, sigmaMap, sigmaDeriv = Props.Invertible(
        "Electrical conductivity at infinite frequency(S/m)"
    )

    chi = Props.PhysicalProperty("Magnetic susceptibility", default=0.0)

    eta, etaMap, etaDeriv = Props.Invertible(
        "Electrical chargeability (V/V), 0 <= eta < 1", default=0.0
    )

    tau, tauMap, tauDeriv = Props.Invertible("Time constant (s)", default=1.0)

    c, cMap, cDeriv = Props.Invertible("Frequency Dependency, 0 < c < 1", default=0.5)

    h, hMap, hDeriv = Props.Invertible("Receiver Height (m), h > 0",)

    def __init__(self, mesh, **kwargs):
        Problem.BaseProblem.__init__(self, mesh, **kwargs)

        # Check input arguments. If self.hankel_filter is not a valid filter,
        # it will set it to the default (key_201_2009).

        ht, htarg = check_hankel(
            "dlf", {"dlf": self.hankel_filter, "pts_per_dec": 0}, 1
        )

        self.fhtfilt = htarg["dlf"]  # Store filter
        self.hankel_pts_per_dec = htarg["pts_per_dec"]  # Store pts_per_dec
        if self.verbose:
            print(">> Use " + self.hankel_filter + " filter for Hankel Transform")

        # if self.hankel_pts_per_dec != 0:
        #     raise NotImplementedError()

    def hz_kernel_vertical_magnetic_dipole(
        self, lamda, f, n_layer, sig, chi, depth, h, z, flag, I, output_type="response"
    ):

        """
            Kernel for vertical magnetic component (Hz) due to
            vertical magnetic diopole (VMD) source in (kx,ky) domain

        """
        u0 = lamda
        coefficient_wavenumber = 1 / (4 * np.pi) * lamda ** 3 / u0

        n_frequency = self.survey.n_frequency
        n_layer = self.survey.n_layer
        n_filter = self.n_filter

        if output_type == "sensitivity_sigma":
            drTE = np.zeros(
                [n_layer, n_frequency, n_filter], dtype=np.complex128, order="F"
            )
            if rte_fortran is None:
                drTE = rTEfunjac(
                    n_layer, f, lamda, sig, chi, depth, self.survey.half_switch
                )
            else:
                rte_fortran.rte_sensitivity(
                    f,
                    lamda,
                    sig,
                    chi,
                    depth,
                    self.survey.half_switch,
                    drTE,
                    n_layer,
                    n_frequency,
                    n_filter,
                )

            kernel = drTE * np.exp(-u0 * (z + h)) * coefficient_wavenumber
        else:
            rTE = np.empty([n_frequency, n_filter], dtype=np.complex128, order="F")
            if rte_fortran is None:
                rTE = rTEfunfwd(
                    n_layer, f, lamda, sig, chi, depth, self.survey.half_switch
                )
            else:
                rte_fortran.rte_forward(
                    f,
                    lamda,
                    sig,
                    chi,
                    depth,
                    self.survey.half_switch,
                    rTE,
                    n_layer,
                    n_frequency,
                    n_filter,
                )

            kernel = rTE * np.exp(-u0 * (z + h)) * coefficient_wavenumber
            if output_type == "sensitivity_height":
                kernel *= -2 * u0

        return kernel * I

        # Note
        # Here only computes secondary field.
        # I am not sure why it does not work if we add primary term.
        # This term can be analytically evaluated, where h = 0.
        #     kernel = (
        #         1./(4*np.pi) *
        #         (np.exp(u0*(z-h))+rTE * np.exp(-u0*(z+h)))*lamda**3/u0
        #     )

    # TODO: make this to take a vector rather than a single frequency
    def hz_kernel_circular_loop(
        self,
        lamda,
        f,
        n_layer,
        sig,
        chi,
        depth,
        h,
        z,
        I,
        a,
        flag,
        output_type="response",
    ):

        """

        Kernel for vertical magnetic component (Hz) at the center
        due to circular loop source in (kx,ky) domain

        .. math::

            H_z = \\frac{Ia}{2} \\int_0^{\\infty} [e^{-u_0|z+h|} +
            \\r_{TE}e^{u_0|z-h|}]
            \\frac{\\lambda^2}{u_0} J_1(\\lambda a)] d \\lambda

        """

        n_frequency = self.survey.n_frequency
        n_layer = self.survey.n_layer
        n_filter = self.n_filter

        w = 2 * np.pi * f
        u0 = lamda
        radius = np.empty([n_frequency, n_filter], order="F")
        radius[:, :] = np.tile(a.reshape([-1, 1]), (1, n_filter))

        coefficient_wavenumber = I * radius * 0.5 * lamda ** 2 / u0

        if output_type == "sensitivity_sigma":
            drTE = np.empty(
                [n_layer, n_frequency, n_filter], dtype=np.complex128, order="F"
            )
            if rte_fortran is None:
                drTE[:, :] = rTEfunjac(
                    n_layer, f, lamda, sig, chi, depth, self.survey.half_switch
                )
            else:
                rte_fortran.rte_sensitivity(
                    f,
                    lamda,
                    sig,
                    chi,
                    depth,
                    self.survey.half_switch,
                    drTE,
                    n_layer,
                    n_frequency,
                    n_filter,
                )

            kernel = drTE * np.exp(-u0 * (z + h)) * coefficient_wavenumber
        else:
            rTE = np.empty([n_frequency, n_filter], dtype=np.complex128, order="F")
            if rte_fortran is None:
                rTE[:, :] = rTEfunfwd(
                    n_layer, f, lamda, sig, chi, depth, self.survey.half_switch
                )
            else:
                rte_fortran.rte_forward(
                    f,
                    lamda,
                    sig,
                    chi,
                    depth,
                    self.survey.half_switch,
                    rTE,
                    n_layer,
                    n_frequency,
                    n_filter,
                )

            if flag == "secondary":
                kernel = rTE * np.exp(-u0 * (z + h)) * coefficient_wavenumber
            else:
                kernel = (
                    rTE
                    * (np.exp(-u0 * (z + h)) + np.exp(u0 * (z - h)))
                    * coefficient_wavenumber
                )

            if output_type == "sensitivity_height":
                kernel *= -2 * u0

        return kernel

    def hz_kernel_horizontal_electric_dipole(
        self, lamda, f, n_layer, sig, chi, depth, h, z, flag, output_type="response"
    ):

        """
            Kernel for vertical magnetic field (Hz) due to
            horizontal electric diopole (HED) source in (kx,ky) domain

        """
        n_frequency = self.survey.n_frequency
        n_layer = self.survey.n_layer
        n_filter = self.n_filter

        u0 = lamda
        coefficient_wavenumber = 1 / (4 * np.pi) * lamda ** 2 / u0

        if output_type == "sensitivity_sigma":
            drTE = np.zeros(
                [n_layer, n_frequency, n_filter], dtype=np.complex128, order="F"
            )
            if rte_fortran is None:
                drTE = rTEfunjac(
                    n_layer, f, lamda, sig, chi, depth, self.survey.half_switch
                )
            else:
                rte_fortran.rte_sensitivity(
                    f,
                    lamda,
                    sig,
                    chi,
                    depth,
                    self.survey.half_switch,
                    drTE,
                    n_layer,
                    n_frequency,
                    n_filter,
                )

            kernel = drTE * np.exp(-u0 * (z + h)) * coefficient_wavenumber
        else:
            rTE = np.empty([n_frequency, n_filter], dtype=np.complex128, order="F")
            if rte_fortran is None:
                rTE = rTEfunfwd(
                    n_layer, f, lamda, sig, chi, depth, self.survey.half_switch
                )
            else:
                rte_fortran.rte_forward(
                    f,
                    lamda,
                    sig,
                    chi,
                    depth,
                    self.survey.half_switch,
                    rTE,
                    n_layer,
                    n_frequency,
                    n_filter,
                )

            kernel = rTE * np.exp(-u0 * (z + h)) * coefficient_wavenumber
            if output_type == "sensitivity_height":
                kernel *= -2 * u0

        return kernel

    # make it as a property?

    def sigma_cole(self):
        """
        Computes Pelton's Cole-Cole conductivity model
        in frequency domain.

        Parameter
        ---------

        n_filter: int
            the number of filter values
        f: ndarray
            frequency (Hz)

        Return
        ------

        sigma_complex: ndarray (n_layer x n_frequency x n_filter)
            Cole-Cole conductivity values at given frequencies

        """
        n_layer = self.survey.n_layer
        n_frequency = self.survey.n_frequency
        n_filter = self.n_filter
        f = self.survey.frequency

        sigma = np.tile(self.sigma.reshape([-1, 1]), (1, n_frequency))
        if np.isscalar(self.eta):
            eta = self.eta
            tau = self.tau
            c = self.c
        else:
            eta = np.tile(self.eta.reshape([-1, 1]), (1, n_frequency))
            tau = np.tile(self.tau.reshape([-1, 1]), (1, n_frequency))
            c = np.tile(self.c.reshape([-1, 1]), (1, n_frequency))

        w = np.tile(2 * np.pi * f, (n_layer, 1))

        sigma_complex = np.empty([n_layer, n_frequency], dtype=np.complex128, order="F")
        sigma_complex[:, :] = sigma - sigma * eta / (
            1 + (1 - eta) * (1j * w * tau) ** c
        )

        sigma_complex_tensor = np.empty(
            [n_layer, n_frequency, n_filter], dtype=np.complex128, order="F"
        )
        sigma_complex_tensor[:, :, :] = np.tile(
            sigma_complex.reshape((n_layer, n_frequency, 1)), (1, 1, n_filter)
        )

        return sigma_complex_tensor

    @property
    def n_filter(self):
        """ Length of filter """
        return self.fhtfilt.base.size

    def forward(self, m, output_type="response"):
        """
            Return Bz or dBzdt
        """

        self.model = m

        n_frequency = self.survey.n_frequency
        flag = self.survey.field_type
        n_layer = self.survey.n_layer
        depth = self.survey.depth
        I = self.survey.I
        n_filter = self.n_filter

        # Get lambd and offset, will depend on pts_per_dec
        if self.survey.src_type == "VMD":
            r = self.survey.offset
        else:
            # a is the radius of the loop
            r = self.survey.a * np.ones(n_frequency)

        # Use function from empymod
        # size of lambd is (n_frequency x n_filter)
        lambd = np.empty([self.survey.frequency.size, n_filter], order="F")
        lambd[:, :], _ = get_dlf_points(self.fhtfilt, r, self.hankel_pts_per_dec)

        # TODO: potentially store
        f = np.empty([self.survey.frequency.size, n_filter], order="F")
        f[:, :] = np.tile(self.survey.frequency.reshape([-1, 1]), (1, n_filter))
        # h is an inversion parameter
        if self.hMap is not None:
            h = self.h
        else:
            h = self.survey.h

        z = h + self.survey.dz

        chi = self.chi

        if np.isscalar(self.chi):
            chi = np.ones_like(self.sigma) * self.chi

        # TODO: potentially store
        sig = self.sigma_cole()

        if output_type == "response":
            # for simulation
            if self.survey.src_type == "VMD":
                hz = self.hz_kernel_vertical_magnetic_dipole(
                    lambd,
                    f,
                    n_layer,
                    sig,
                    chi,
                    depth,
                    h,
                    z,
                    flag,
                    I,
                    output_type=output_type,
                )

                # kernels for each bessel function
                # (j0, j1, j2)
                PJ = (hz, None, None)  # PJ0

            elif self.survey.src_type == "CircularLoop":
                hz = self.hz_kernel_circular_loop(
                    lambd,
                    f,
                    n_layer,
                    sig,
                    chi,
                    depth,
                    h,
                    z,
                    I,
                    r,
                    flag,
                    output_type=output_type,
                )

                # kernels for each bessel function
                # (j0, j1, j2)
                PJ = (None, hz, None)  # PJ1

            # TODO: This has not implemented yet!
            elif self.survey.src_type == "piecewise_line":
                # Need to compute y
                hz = self.hz_kernel_horizontal_electric_dipole(
                    lambd,
                    f,
                    n_layer,
                    sig,
                    chi,
                    depth,
                    h,
                    z,
                    I,
                    r,
                    flag,
                    output_type=output_type,
                )
                # kernels for each bessel function
                # (j0, j1, j2)
                PJ = (None, hz, None)  # PJ1

            else:
                raise Exception("Src options are only VMD or CircularLoop!!")

        elif output_type == "sensitivity_sigma":

            # for simulation
            if self.survey.src_type == "VMD":
                hz = self.hz_kernel_vertical_magnetic_dipole(
                    lambd,
                    f,
                    n_layer,
                    sig,
                    chi,
                    depth,
                    h,
                    z,
                    flag,
                    I,
                    output_type=output_type,
                )

                PJ = (hz, None, None)  # PJ0

            elif self.survey.src_type == "CircularLoop":

                hz = self.hz_kernel_circular_loop(
                    lambd,
                    f,
                    n_layer,
                    sig,
                    chi,
                    depth,
                    h,
                    z,
                    I,
                    r,
                    flag,
                    output_type=output_type,
                )

                PJ = (None, hz, None)  # PJ1

            else:
                raise Exception("Src options are only VMD or CircularLoop!!")

            r = np.tile(r, (n_layer, 1))

        elif output_type == "sensitivity_height":

            # for simulation
            if self.survey.src_type == "VMD":
                hz = self.hz_kernel_vertical_magnetic_dipole(
                    lambd,
                    f,
                    n_layer,
                    sig,
                    chi,
                    depth,
                    h,
                    z,
                    flag,
                    I,
                    output_type=output_type,
                )

                PJ = (hz, None, None)  # PJ0

            elif self.survey.src_type == "CircularLoop":

                hz = self.hz_kernel_circular_loop(
                    lambd,
                    f,
                    n_layer,
                    sig,
                    chi,
                    depth,
                    h,
                    z,
                    I,
                    r,
                    flag,
                    output_type=output_type,
                )

                PJ = (None, hz, None)  # PJ1

            else:
                raise Exception("Src options are only VMD or CircularLoop!!")

        # Carry out Hankel DLF
        # ab=66 => 33 (vertical magnetic src and rec)
        # For response
        # HzFHT size = (n_frequency,)
        # For sensitivity
        # HzFHT size = (n_layer, n_frequency)

        HzFHT = dlf(
            PJ, lambd, r, self.fhtfilt, self.hankel_pts_per_dec, ang_fact=None, ab=33
        )

        if output_type == "sensitivity_sigma":
            return HzFHT.T

        return HzFHT

    # @profile
    def fields(self, m):
        f = self.forward(m, output_type="response")
        self.survey._pred = Utils.mkvc(self.survey.projectFields(f))
        return f

    def getJ_height(self, m, f=None):
        """

        """
        if self.hMap is None:
            return Utils.Zero()

        if self._Jmatrix_height is not None:
            return self._Jmatrix_height
        else:

            if self.verbose:
                print(">> Compute J height ")

            dudz = self.forward(m, output_type="sensitivity_height")

            self._Jmatrix_height = (self.survey.projectFields(dudz)).reshape([-1, 1])

            return self._Jmatrix_height

    # @profile
    def getJ_sigma(self, m, f=None):

        if self.sigmaMap is None:
            return Utils.Zero()

        if self._Jmatrix_sigma is not None:
            return self._Jmatrix_sigma
        else:

            if self.verbose:
                print(">> Compute J sigma")

            dudsig = self.forward(m, output_type="sensitivity_sigma")

            self._Jmatrix_sigma = self.survey.projectFields(dudsig)
            if self._Jmatrix_sigma.ndim == 1:
                self._Jmatrix_sigma = self._Jmatrix_sigma.reshape([-1, 1])
            return self._Jmatrix_sigma

    def getJ(self, m, f=None):
        return (
            self.getJ_sigma(m, f=f) * self.sigmaDeriv
            + self.getJ_height(m, f=f) * self.hDeriv
        )

    def Jvec(self, m, v, f=None):
        """
            Computing Jacobian^T multiplied by vector.
        """

        J_sigma = self.getJ_sigma(m, f=f)
        J_height = self.getJ_height(m, f=f)
        Jv = np.dot(J_sigma, self.sigmaMap.deriv(m, v))
        if self.hMap is not None:
            Jv += np.dot(J_height, self.hMap.deriv(m, v))
        return Jv

    def Jtvec(self, m, v, f=None):
        """
            Computing Jacobian^T multiplied by vector.
        """

        J_sigma = self.getJ_sigma(m, f=f)
        J_height = self.getJ_height(m, f=f)
        Jtv = self.sigmaDeriv.T * np.dot(J_sigma.T, v)
        if self.hMap is not None:
            Jtv += self.hDeriv.T * np.dot(J_height.T, v)
        return Jtv

    @property
    def deleteTheseOnModelUpdate(self):
        toDelete = []
        if self.fix_Jmatrix is False:
            if self._Jmatrix_sigma is not None:
                toDelete += ["_Jmatrix_sigma"]
            if self._Jmatrix_height is not None:
                toDelete += ["_Jmatrix_height"]
        return toDelete

    def depth_of_investigation_christiansen_2012(self, std, thres_hold=0.8):
        pred = self.survey._pred.copy()
        delta_d = std * np.log(abs(self.survey.dobs))
        J = self.getJ(self.model)
        J_sum = abs(Utils.sdiag(1 / delta_d / pred) * J).sum(axis=0)
        S = np.cumsum(J_sum[::-1])[::-1]
        active = S - thres_hold > 0.0
        doi = abs(self.survey.depth[active]).max()
        return doi, active

    def get_threshold(self, uncert):
        _, active = self.depth_of_investigation(uncert)
        JtJdiag = self.get_JtJdiag(uncert)
        delta = JtJdiag[active].min()
        return delta

    def get_JtJdiag(self, uncert):
        J = self.getJ(self.model)
        JtJdiag = (np.power((Utils.sdiag(1.0 / uncert) * J), 2)).sum(axis=0)
        return JtJdiag


if __name__ == "__main__":
    main()



import os
from shutil import copyfile
import time
from ipywidgets import Checkbox, Text, VBox, Label, ToggleButton, Widget
from geoh5py.workspace import Workspace


class BaseApplication:
    """
    Base class for geoapps applications
    """

    def __init__(self, **kwargs):

        self._h5file = "Analyst.geoh5"
        self._live_link = Checkbox(
            description="GA Pro - Live link", value=False, indent=False
        )

        def live_link_choice(_):
            self.live_link_choice()

        self._live_link.observe(live_link_choice)

        self._live_link_path = Text(description="", value="", disabled=True,)

        self._trigger = ToggleButton(
            value=False,
            description="Compute",
            button_style="danger",
            tooltip="Run computation",
            icon="check",
        )
        self.trigger_widget = VBox(
            [
                self.trigger,
                self.live_link,
                Label("Monitoring folder"),
                self.live_link_path,
            ]
        )

        for obj in self.__dict__:
            if hasattr(getattr(self, obj), "style"):
                getattr(self, obj).style = {"description_width": "initial"}

        for key, value in kwargs.items():
            if getattr(self, "_" + key, None) is not None:
                try:
                    if isinstance(getattr(self, "_" + key), Widget):
                        getattr(self, "_" + key).value = value
                    else:
                        setattr(self, "_" + key, value)
                except:
                    pass

        if self.h5file is not None:
            self.live_link_path.value = os.path.join(
                os.path.abspath(os.path.dirname(self.h5file)), "Temp"
            )

    def live_link_output(self, entity, data={}):
        """
        Create a temporary geoh5 file in the monitoring folder and export entity for update.

        :param :obj:`geoh5py.Entity`: Entity to be updated
        :param data: `dict` of values to be added as data {"name": values}
        """
        if not os.path.exists(self.live_link_path.value):
            os.mkdir(self.live_link_path.value)

        temp_geoh5 = os.path.join(
            self.live_link_path.value, f"temp{time.time():.3f}.geoh5"
        )
        temp_workspace = Workspace(temp_geoh5)

        for key, value in data.items():
            entity.add_data({key: {"values": value}})

        entity.copy(parent=temp_workspace)

    def live_link_choice(self):
        """
        Enable the monitoring folder
        """
        if self.live_link.value:
            self.live_link_path.disabled = False
        else:
            self.live_link_path.disabled = True

    def widget(self):
        ...

    @property
    def h5file(self):
        """
        :obj:`str`: Target geoh5 project file.
        """
        return self._h5file

    @property
    def live_link(self):
        """
        :obj:`ipywidgets.Checkbox`: Activate the live link between an application and Geoscience ANALYST
        """
        return self._live_link

    @property
    def live_link_path(self):
        """
        :obj:`ipywidgets.Text`: Path for the monitoring folder to be copied to Geoscience ANALYST preferences.
        """
        return self._live_link_path

    @property
    def trigger(self):
        """
        :obj:`ipywidgets.ToggleButton`: Trigger some computation and output.
        """
        return self._trigger

    @property
    def workspace(self):
        """
        Target geoh5py workspace
        """
        if (
            getattr(self, "_workspace", None) is None
            and getattr(self, "_h5file", None) is not None
        ):
            self._workspace = Workspace(self.h5file)
        return self._workspace


def working_copy(**kwargs):
    """
    Create a copy of the geoh5 project and remove "working_copy" from list
    of arguments for future use
    """
    if (
        "h5file" in kwargs.keys()
        and "working_copy" in kwargs.keys()
        and kwargs["working_copy"]
    ):
        kwargs["h5file"] = copyfile(
            kwargs["h5file"], kwargs["h5file"][:-6] + "_work.geoh5"
        )

        del kwargs["working_copy"]

    return kwargs



from .. import Utils
from SimPEG.EM.Base import BaseEMProblem
from SimPEG.EM.Static.DC.FieldsDC import FieldsDC, Fields_CC, Fields_N
import numpy as np
from SimPEG.Utils import Zero
from SimPEG.EM.Static.DC import getxBCyBC_CC
from .SurveyIP import Survey
from .. import Props


class BaseIPProblem(BaseEMProblem):

    sigma = Props.PhysicalProperty("Electrical conductivity (S/m)")

    rho = Props.PhysicalProperty("Electrical resistivity (Ohm m)")

    Props.Reciprocal(sigma, rho)

    eta, etaMap, etaDeriv = Props.Invertible("Electrical Chargeability")

    surveyPair = Survey
    fieldsPair = FieldsDC
    Ainv = None
    f = None
    Ainv = None

    def fields(self, m):
        if m is not None:
            self.model = m
        if self.f is None:
            self.f = self.fieldsPair(self.mesh, self.survey)
            if self.Ainv is None:
                A = self.getA()
                self.Ainv = self.Solver(A, **self.solverOpts)
            RHS = self.getRHS()
            u = self.Ainv * RHS
            Srcs = self.survey.srcList
            self.f[Srcs, self._solutionType] = u
        return self.f

    def Jvec(self, m, v, f=None):

        if f is None:
            f = self.fields(m)

        self.model = m

        Jv = []
        A = self.getA()

        for src in self.survey.srcList:
            u_src = f[src, self._solutionType]  # solution vector
            dA_dm_v = self.getADeriv(u_src, v)
            dRHS_dm_v = self.getRHSDeriv(src, v)
            du_dm_v = self.Ainv * (-dA_dm_v + dRHS_dm_v)

            for rx in src.rxList:
                df_dmFun = getattr(f, f"_{rx.projField!s}Deriv", None)
                df_dm_v = df_dmFun(src, du_dm_v, v, adjoint=False)
                # Jv[src, rx] = rx.evalDeriv(src, self.mesh, f, df_dm_v)
                Jv.append(rx.evalDeriv(src, self.mesh, f, df_dm_v))
        # Conductivity (d u / d log sigma)
        if self._formulation == "EB":
            # return -Utils.mkvc(Jv)
            return -np.hstack(Jv)
        # Conductivity (d u / d log rho)
        if self._formulation == "HJ":
            # return Utils.mkvc(Jv)
            return np.hstack(Jv)

    def Jtvec(self, m, v, f=None):
        if f is None:
            f = self.fields(m)

        self.model = m

        # Ensure v is a data object.
        if not isinstance(v, self.dataPair):
            v = self.dataPair(self.survey, v)

        Jtv = np.zeros(m.size)
        AT = self.getA()

        for src in self.survey.srcList:
            u_src = f[src, self._solutionType]
            for rx in src.rxList:
                PTv = rx.evalDeriv(
                    src, self.mesh, f, v[src, rx], adjoint=True
                )  # wrt f, need possibility wrt m
                df_duTFun = getattr(f, f"_{rx.projField!s}Deriv", None)
                df_duT, df_dmT = df_duTFun(src, None, PTv, adjoint=True)
                ATinvdf_duT = self.Ainv * df_duT
                dA_dmT = self.getADeriv(u_src, ATinvdf_duT, adjoint=True)
                dRHS_dmT = self.getRHSDeriv(src, ATinvdf_duT, adjoint=True)
                du_dmT = -dA_dmT + dRHS_dmT
                Jtv += (df_dmT + du_dmT).astype(float)
        # Conductivity ((d u / d log sigma).T)
        if self._formulation == "EB":
            return -Utils.mkvc(Jtv)
        # Conductivity ((d u / d log rho).T)
        if self._formulation == "HJ":
            return Utils.mkvc(Jtv)

    def getSourceTerm(self):
        """
        takes concept of source and turns it into a matrix
        """
        """
        Evaluates the sources, and puts them in matrix form

        :rtype: (numpy.ndarray, numpy.ndarray)
        :return: q (nC or nN, nSrc)
        """

        Srcs = self.survey.srcList

        if self._formulation == "EB":
            n = self.mesh.nN
            # return NotImplementedError

        elif self._formulation == "HJ":
            n = self.mesh.nC

        q = np.zeros((n, len(Srcs)))

        for i, src in enumerate(Srcs):
            q[:, i] = src.eval(self)
        return q

    @property
    def deleteTheseOnModelUpdate(self):
        toDelete = []
        return toDelete

    # assume log rho or log cond
    @property
    def MeSigma(self):
        """
            Edge inner product matrix for \\(\\sigma\\).
            Used in the E-B formulation
        """
        if getattr(self, "_MeSigma", None) is None:
            self._MeSigma = self.mesh.getEdgeInnerProduct(self.sigma)
        return self._MeSigma

    @property
    def MfRhoI(self):
        """
            Inverse of :code:`MfRho`
        """
        if getattr(self, "_MfRhoI", None) is None:
            self._MfRhoI = self.mesh.getFaceInnerProduct(self.rho, invMat=True)
        return self._MfRhoI

    def MfRhoIDeriv(self, u):
        """
            Derivative of :code:`MfRhoI` with respect to the model.
        """

        dMfRhoI_dI = -self.MfRhoI ** 2
        dMf_drho = self.mesh.getFaceInnerProductDeriv(self.rho)(u)
        drho_dlogrho = Utils.sdiag(self.rho) * self.etaDeriv
        return dMfRhoI_dI * (dMf_drho * drho_dlogrho)

    # TODO: This should take a vector
    def MeSigmaDeriv(self, u):
        """
            Derivative of MeSigma with respect to the model
        """
        dsigma_dlogsigma = Utils.sdiag(self.sigma) * self.etaDeriv
        return self.mesh.getEdgeInnerProductDeriv(self.sigma)(u) * dsigma_dlogsigma


class Problem3D_CC(BaseIPProblem):

    _solutionType = "phiSolution"
    _formulation = "HJ"  # CC potentials means J is on faces
    fieldsPair = Fields_CC

    def __init__(self, mesh, **kwargs):
        BaseIPProblem.__init__(self, mesh, **kwargs)
        self.setBC()

    def getA(self):
        """

        Make the A matrix for the cell centered DC resistivity problem

        A = D MfRhoI G

        """

        D = self.Div
        G = self.Grad
        MfRhoI = self.MfRhoI
        A = D * MfRhoI * G

        # I think we should deprecate this for DC problem.
        # if self._makeASymmetric is True:
        #     return V.T * A
        return A

    def getADeriv(self, u, v, adjoint=False):

        D = self.Div
        G = self.Grad
        MfRhoIDeriv = self.MfRhoIDeriv

        if adjoint:
            # if self._makeASymmetric is True:
            #     v = V * v
            return (MfRhoIDeriv(G * u).T) * (D.T * v)

        # I think we should deprecate this for DC problem.
        # if self._makeASymmetric is True:
        #     return V.T * ( D * ( MfRhoIDeriv( D.T * ( V * u ) ) * v ) )
        return D * (MfRhoIDeriv(G * u) * v)

    def getRHS(self):
        """
        RHS for the DC problem

        q
        """

        RHS = self.getSourceTerm()

        # I think we should deprecate this for DC problem.
        # if self._makeASymmetric is True:
        #     return self.Vol.T * RHS

        return RHS

    def getRHSDeriv(self, src, v, adjoint=False):
        """
        Derivative of the right hand side with respect to the model
        """
        # TODO: add qDeriv for RHS depending on m
        # qDeriv = src.evalDeriv(self, adjoint=adjoint)
        # return qDeriv
        return Zero()

    def setBC(self):
        if self.mesh.dim == 3:
            fxm, fxp, fym, fyp, fzm, fzp = self.mesh.faceBoundaryInd
            gBFxm = self.mesh.gridFx[fxm, :]
            gBFxp = self.mesh.gridFx[fxp, :]
            gBFym = self.mesh.gridFy[fym, :]
            gBFyp = self.mesh.gridFy[fyp, :]
            gBFzm = self.mesh.gridFz[fzm, :]
            gBFzp = self.mesh.gridFz[fzp, :]

            # Setup Mixed B.C (alpha, beta, gamma)
            temp_xm, temp_xp = np.ones_like(gBFxm[:, 0]), np.ones_like(gBFxp[:, 0])
            temp_ym, temp_yp = np.ones_like(gBFym[:, 1]), np.ones_like(gBFyp[:, 1])
            temp_zm, temp_zp = np.ones_like(gBFzm[:, 2]), np.ones_like(gBFzp[:, 2])

            alpha_xm, alpha_xp = temp_xm * 0.0, temp_xp * 0.0
            alpha_ym, alpha_yp = temp_ym * 0.0, temp_yp * 0.0
            alpha_zm, alpha_zp = temp_zm * 0.0, temp_zp * 0.0

            beta_xm, beta_xp = temp_xm, temp_xp
            beta_ym, beta_yp = temp_ym, temp_yp
            beta_zm, beta_zp = temp_zm, temp_zp

            gamma_xm, gamma_xp = temp_xm * 0.0, temp_xp * 0.0
            gamma_ym, gamma_yp = temp_ym * 0.0, temp_yp * 0.0
            gamma_zm, gamma_zp = temp_zm * 0.0, temp_zp * 0.0

            alpha = [alpha_xm, alpha_xp, alpha_ym, alpha_yp, alpha_zm, alpha_zp]
            beta = [beta_xm, beta_xp, beta_ym, beta_yp, beta_zm, beta_zp]
            gamma = [gamma_xm, gamma_xp, gamma_ym, gamma_yp, gamma_zm, gamma_zp]

        elif self.mesh.dim == 2:

            fxm, fxp, fym, fyp = self.mesh.faceBoundaryInd
            gBFxm = self.mesh.gridFx[fxm, :]
            gBFxp = self.mesh.gridFx[fxp, :]
            gBFym = self.mesh.gridFy[fym, :]
            gBFyp = self.mesh.gridFy[fyp, :]

            # Setup Mixed B.C (alpha, beta, gamma)
            temp_xm, temp_xp = np.ones_like(gBFxm[:, 0]), np.ones_like(gBFxp[:, 0])
            temp_ym, temp_yp = np.ones_like(gBFym[:, 1]), np.ones_like(gBFyp[:, 1])

            alpha_xm, alpha_xp = temp_xm * 0.0, temp_xp * 0.0
            alpha_ym, alpha_yp = temp_ym * 0.0, temp_yp * 0.0

            beta_xm, beta_xp = temp_xm, temp_xp
            beta_ym, beta_yp = temp_ym, temp_yp

            gamma_xm, gamma_xp = temp_xm * 0.0, temp_xp * 0.0
            gamma_ym, gamma_yp = temp_ym * 0.0, temp_yp * 0.0

            alpha = [alpha_xm, alpha_xp, alpha_ym, alpha_yp]
            beta = [beta_xm, beta_xp, beta_ym, beta_yp]
            gamma = [gamma_xm, gamma_xp, gamma_ym, gamma_yp]

        x_BC, y_BC = getxBCyBC_CC(self.mesh, alpha, beta, gamma)
        V = self.Vol
        self.Div = V * self.mesh.faceDiv
        P_BC, B = self.mesh.getBCProjWF_simple()
        M = B * self.mesh.aveCC2F
        self.Grad = self.Div.T - P_BC * Utils.sdiag(y_BC) * M


class Problem3D_N(BaseIPProblem):

    _solutionType = "phiSolution"
    _formulation = "EB"  # N potentials means B is on faces
    fieldsPair = Fields_N

    def __init__(self, mesh, **kwargs):
        BaseIPProblem.__init__(self, mesh, **kwargs)

    def getA(self):
        """

        Make the A matrix for the cell centered DC resistivity problem

        A = G.T MeSigma G

        """

        MeSigma = self.MeSigma
        Grad = self.mesh.nodalGrad
        A = Grad.T * MeSigma * Grad

        # Handling Null space of A
        A[0, 0] = A[0, 0] + 1.0

        return A

    def getADeriv(self, u, v, adjoint=False):
        """
            Product of the derivative of our system matrix with
            respect to the model and a vector
        """
        Grad = self.mesh.nodalGrad
        if not adjoint:
            return Grad.T * (self.MeSigmaDeriv(Grad * u) * v)
        elif adjoint:
            return self.MeSigmaDeriv(Grad * u).T * (Grad * v)

    def getRHS(self):
        """
        RHS for the DC problem

        q
        """

        RHS = self.getSourceTerm()
        return RHS

    def getRHSDeriv(self, src, v, adjoint=False):
        """
        Derivative of the right hand side with respect to the model
        """
        # TODO: add qDeriv for RHS depending on m
        # qDeriv = src.evalDeriv(self, adjoint=adjoint)
        # return qDeriv
        return Zero()



import numpy as np
import scipy.sparse as sp
import SimPEG
from .. import Utils
from SimPEG.EM.Utils import omega
from SimPEG.Utils import Zero, Identity


class FieldsTDEM(SimPEG.Problem.TimeFields):
    """

    Fancy Field Storage for a TDEM survey. Only one field type is stored for
    each problem, the rest are computed. The fields obejct acts like an array
    and is indexed by

    .. code-block:: python

        f = problem.fields(m)
        e = f[srcList,'e']
        b = f[srcList,'b']

    If accessing all sources for a given field, use the :code:`:`

    .. code-block:: python

        f = problem.fields(m)
        e = f[:,'e']
        b = f[:,'b']

    The array returned will be size (nE or nF, nSrcs :math:`\\times`
    nFrequencies)
    """

    knownFields = {}
    dtype = float

    def _GLoc(self, fieldType):
        """Grid location of the fieldType"""
        return self.aliasFields[fieldType][1]

    def _eDeriv(self, tInd, src, dun_dm_v, v, adjoint=False):
        if adjoint is True:
            return (
                self._eDeriv_u(tInd, src, v, adjoint),
                self._eDeriv_m(tInd, src, v, adjoint),
            )
        return self._eDeriv_u(tInd, src, dun_dm_v) + self._eDeriv_m(tInd, src, v)

    def _bDeriv(self, tInd, src, dun_dm_v, v, adjoint=False):
        if adjoint is True:
            return (
                self._bDeriv_u(tInd, src, v, adjoint),
                self._bDeriv_m(tInd, src, v, adjoint),
            )
        return self._bDeriv_u(tInd, src, dun_dm_v) + self._bDeriv_m(tInd, src, v)

    def _dbdtDeriv(self, tInd, src, dun_dm_v, v, adjoint=False):
        if adjoint is True:
            return (
                self._dbdtDeriv_u(tInd, src, v, adjoint),
                self._dbdtDeriv_m(tInd, src, v, adjoint),
            )
        return self._dbdtDeriv_u(tInd, src, dun_dm_v) + self._dbdtDeriv_m(tInd, src, v)

    def _hDeriv(self, tInd, src, dun_dm_v, v, adjoint=False):
        if adjoint is True:
            return (
                self._hDeriv_u(tInd, src, v, adjoint),
                self._hDeriv_m(tInd, src, v, adjoint),
            )
        return self._hDeriv_u(tInd, src, dun_dm_v) + self._hDeriv_m(tInd, src, v)

    def _dhdtDeriv(self, tInd, src, dun_dm_v, v, adjoint=False):
        if adjoint is True:
            return (
                self._dhdtDeriv_u(tInd, src, v, adjoint),
                self._dhdtDeriv_m(tInd, src, v, adjoint),
            )
        return self._dhdtDeriv_u(tInd, src, dun_dm_v) + self._dhdtDeriv_m(tInd, src, v)

    def _jDeriv(self, tInd, src, dun_dm_v, v, adjoint=False):
        if adjoint is True:
            return (
                self._jDeriv_u(tInd, src, v, adjoint),
                self._jDeriv_m(tInd, src, v, adjoint),
            )
        return self._jDeriv_u(tInd, src, dun_dm_v) + self._jDeriv_m(tInd, src, v)


class Fields_Derivs(FieldsTDEM):
    """
        A fields object for satshing derivs
    """

    knownFields = {
        "bDeriv": "F",
        "eDeriv": "E",
        "hDeriv": "E",
        "jDeriv": "F",
        "dbdtDeriv": "F",
        "dhdtDeriv": "E",
    }


class Fields3D_b(FieldsTDEM):
    """Field Storage for a TDEM survey."""

    knownFields = {"bSolution": "F"}
    aliasFields = {
        "b": ["bSolution", "F", "_b"],
        "e": ["bSolution", "E", "_e"],
        "dbdt": ["bSolution", "F", "_dbdt"],
    }

    def startup(self):
        self._MeSigmaI = self.survey.prob.MeSigmaI
        self._MeSigmaIDeriv = self.survey.prob.MeSigmaIDeriv
        self._edgeCurl = self.survey.prob.mesh.edgeCurl
        self._MfMui = self.survey.prob.MfMui
        self._timeMesh = self.survey.prob.timeMesh

    def _TLoc(self, fieldType):
        if fieldType in ["e", "b"]:
            return "N"
        elif fieldType == "dbdt":
            return "N"

    def _b(self, bSolution, srcList, tInd):
        return bSolution

    def _bDeriv_u(self, tInd, src, dun_dm_v, adjoint=False):
        return dun_dm_v

    def _bDeriv_m(self, tInd, src, v, adjoint=False):
        return Zero()

    def _dbdt(self, bSolution, srcList, tInd):
        # self._timeMesh.faceDiv
        dbdt = -self._edgeCurl * self._e(bSolution, srcList, tInd)
        for i, src in enumerate(srcList):
            s_m = src.s_m(self.survey.prob, self.survey.prob.times[tInd])
            dbdt[:, i] = dbdt[:, i] + s_m
        return dbdt

    def _dbdtDeriv_u(self, tInd, src, dun_dm_v, adjoint=False):
        if adjoint is True:
            return -self._eDeriv_u(tInd, src, self._edgeCurl.T * dun_dm_v, adjoint)
        return -(self._edgeCurl * self._eDeriv_u(tInd, src, dun_dm_v))

    def _dbdtDeriv_m(self, tInd, src, v, adjoint=False):
        if adjoint is True:
            return -(self._eDeriv_m(tInd, src, self._edgeCurl.T * v, adjoint))
        return -(
            self._edgeCurl * self._eDeriv_m(tInd, src, v)
        )  # + src.s_mDeriv() assuming src doesn't have deriv for now

    def _e(self, bSolution, srcList, tInd):
        e = self._MeSigmaI * (self._edgeCurl.T * (self._MfMui * bSolution))
        for i, src in enumerate(srcList):
            s_e = src.s_e(self.survey.prob, self.survey.prob.times[tInd])
            e[:, i] = e[:, i] - self._MeSigmaI * s_e
        return e

    def _eDeriv_u(self, tInd, src, dun_dm_v, adjoint=False):
        if adjoint is True:
            return self._MfMui.T * (self._edgeCurl * (self._MeSigmaI.T * dun_dm_v))
        return self._MeSigmaI * (self._edgeCurl.T * (self._MfMui * dun_dm_v))

    def _eDeriv_m(self, tInd, src, v, adjoint=False):
        _, s_e = src.eval(self.survey.prob, self.survey.prob.times[tInd])
        bSolution = self[[src], "bSolution", tInd].flatten()

        _, s_eDeriv = src.evalDeriv(self.survey.prob.times[tInd], self, adjoint=adjoint)

        if adjoint is True:
            return self._MeSigmaIDeriv(
                -s_e + self._edgeCurl.T * (self._MfMui * bSolution)
            ).T * v - s_eDeriv(self._MeSigmaI.T * v)

        return self._MeSigmaIDeriv(
            -s_e + self._edgeCurl.T * (self._MfMui * bSolution)
        ) * v - self._MeSigmaI * s_eDeriv(v)


class Fields3D_e(FieldsTDEM):
    """Fancy Field Storage for a TDEM survey."""

    knownFields = {"eSolution": "E"}
    aliasFields = {
        "e": ["eSolution", "E", "_e"],
        "b": ["eSolution", "F", "_b"],
        "dbdt": ["eSolution", "F", "_dbdt"],
    }

    def startup(self):
        self._MeSigmaI = self.survey.prob.MeSigmaI
        self._MeSigmaIDeriv = self.survey.prob.MeSigmaIDeriv
        self._edgeCurl = self.survey.prob.mesh.edgeCurl
        self._MfMui = self.survey.prob.MfMui
        self._times = self.survey.prob.times

    def _TLoc(self, fieldType):
        if fieldType in ["e", "b"]:
            return "N"
        elif fieldType == "dbdt":
            return "N"
        else:
            raise NotImplementedError

    def _e(self, eSolution, srcList, tInd):
        return eSolution

    def _eDeriv_u(self, tInd, src, dun_dm_v, adjoint=False):
        return dun_dm_v

    def _eDeriv_m(self, tInd, src, v, adjoint=False):
        return Zero()

    def _dbdt(self, eSolution, srcList, tInd):
        s_m = np.zeros((self.mesh.nF, len(srcList)))
        for i, src in enumerate(srcList):
            s_m_src = src.s_m(self.survey.prob, self._times[tInd])
            s_m[:, i] = s_m[:, i] + s_m_src
        return s_m - self._edgeCurl * eSolution

    def _dbdtDeriv_u(self, tInd, src, dun_dm_v, adjoint=False):
        if adjoint:
            return -self._edgeCurl.T * dun_dm_v
        return -self._edgeCurl * dun_dm_v

    def _dbdtDeriv_m(self, tInd, src, v, adjoint=False):
        # s_mDeriv = src.s_mDeriv(
        #     self._times[tInd], self, adjoint=adjoint
        # )
        return Utils.Zero()  # assumes source doesn't depend on model

    def _b(self, eSolution, srcList, tInd):
        """
        Integrate _db_dt using rectangles
        """
        raise NotImplementedError("To obtain b-fields, please use Problem3D_b")
        # dbdt = self._dbdt(eSolution, srcList, tInd)
        # dt = self.survey.prob.timeMesh.hx
        # # assume widths of "ghost cells" same on either end
        # dtn = np.hstack([dt[0], 0.5*(dt[1:] + dt[:-1]), dt[-1]])
        # return dtn[tInd] * dbdt
        # # raise NotImplementedError


class Fields3D_h(FieldsTDEM):
    """Fancy Field Storage for a TDEM survey."""

    knownFields = {"hSolution": "E"}
    aliasFields = {
        "h": ["hSolution", "E", "_h"],
        "dhdt": ["hSolution", "E", "_dhdt"],
        "j": ["hSolution", "F", "_j"],
    }

    def startup(self):
        self._edgeCurl = self.survey.prob.mesh.edgeCurl
        self._times = self.survey.prob.times
        self._MeMuI = self.survey.prob.MeMuI
        self._MfRho = self.survey.prob.MfRho
        self._MfRhoDeriv = self.survey.prob.MfRhoDeriv

    def _TLoc(self, fieldType):
        # if fieldType in ['h', 'j']:
        return "N"
        # else:
        #     raise NotImplementedError

    def _h(self, hSolution, srcList, tInd):
        return hSolution

    def _hDeriv_u(self, tInd, src, dun_dm_v, adjoint=False):
        return dun_dm_v

    def _hDeriv_m(self, tInd, src, v, adjoint=False):
        return Zero()

    def _dhdt(self, hSolution, srcList, tInd):
        C = self._edgeCurl
        MeMuI = self._MeMuI
        MfRho = self._MfRho

        dhdt = -MeMuI * (C.T * (MfRho * (C * hSolution)))

        for i, src in enumerate(srcList):
            s_m, s_e = src.eval(self.survey.prob, self._times[tInd])
            dhdt[:, i] = MeMuI * (C.T * MfRho * s_e + s_m) + dhdt[:, i]
        return dhdt

    def _dhdtDeriv_u(self, tInd, src, dun_dm_v, adjoint=False):
        C = self._edgeCurl
        MeMuI = self._MeMuI
        MfRho = self._MfRho

        if adjoint:
            return -C.T * (MfRho.T * (C * (MeMuI * dun_dm_v)))
        return -MeMuI * (C.T * (MfRho * (C * dun_dm_v)))

    def _dhdtDeriv_m(self, tInd, src, v, adjoint=False):
        C = self._edgeCurl
        MeMuI = self._MeMuI
        MfRho = self._MfRho
        MfRhoDeriv = self._MfRhoDeriv

        hSolution = self[[src], "hSolution", tInd].flatten()
        s_e = src.s_e(self.survey.prob, self._times[tInd])

        if adjoint:
            return -MfRhoDeriv(C * hSolution - s_e).T * (C * (MeMuI * v))
        return -MeMuI * (C.T * (MfRhoDeriv(C * hSolution - s_e) * v))

    def _j(self, hSolution, srcList, tInd):
        s_e = np.zeros((self.mesh.nF, len(srcList)))
        for i, src in enumerate(srcList):
            s_e_src = src.s_e(self.survey.prob, self._times[tInd])
            s_e[:, i] = s_e[:, i] + s_e_src

        return self._edgeCurl * hSolution - s_e

    def _jDeriv_u(self, tInd, src, dun_dm_v, adjoint=False):
        if adjoint:
            return self._edgeCurl.T * dun_dm_v
        return self._edgeCurl * dun_dm_v

    def _jDeriv_m(self, tInd, src, v, adjoint=False):
        return Zero()  # assumes the source doesn't depend on the model


class Fields3D_j(FieldsTDEM):
    """Fancy Field Storage for a TDEM survey."""

    knownFields = {"jSolution": "F"}
    aliasFields = {
        "dhdt": ["jSolution", "E", "_dhdt"],
        "j": ["jSolution", "F", "_j"],
    }

    def startup(self):
        self._edgeCurl = self.survey.prob.mesh.edgeCurl
        self._times = self.survey.prob.times
        self._MeMuI = self.survey.prob.MeMuI
        self._MfRho = self.survey.prob.MfRho
        self._MfRhoDeriv = self.survey.prob.MfRhoDeriv

    def _TLoc(self, fieldType):
        # if fieldType in ['h', 'j']:
        return "N"

    def _j(self, jSolution, srcList, tInd):
        return jSolution

    def _jDeriv_u(self, tInd, src, dun_dm_v, adjoint=False):
        return dun_dm_v

    def _jDeriv_m(self, tInd, src, v, adjoint=False):
        return Zero()

    def _h(self, jSolution, srcList, tInd):
        raise NotImplementedError("Please use Problem3D_h to get h-fields")

    def _dhdt(self, jSolution, srcList, tInd):
        C = self._edgeCurl
        MfRho = self._MfRho
        MeMuI = self._MeMuI

        dhdt = -MeMuI * (C.T * (MfRho * jSolution))
        for i, src in enumerate(srcList):
            s_m = src.s_m(self.survey.prob, self.survey.prob.times[tInd])
            dhdt[:, i] = MeMuI * s_m + dhdt[:, i]

        return dhdt

    def _dhdtDeriv_u(self, tInd, src, dun_dm_v, adjoint=False):
        C = self._edgeCurl
        MfRho = self._MfRho
        MeMuI = self._MeMuI

        if adjoint is True:
            return -MfRho.T * (C * (MeMuI.T * dun_dm_v))
        return -MeMuI * (C.T * (MfRho * dun_dm_v))

    def _dhdtDeriv_m(self, tInd, src, v, adjoint=False):
        jSolution = self[[src], "jSolution", tInd].flatten()
        MfRhoDeriv = self._MfRhoDeriv(jSolution)
        C = self._edgeCurl
        MeMuI = self._MeMuI

        if adjoint is True:
            return -MfRhoDeriv.T * (C * (MeMuI * v))
        return -MeMuI * (C.T * (MfRhoDeriv * v))



import os
import re
import discretize
import ipywidgets as widgets
import matplotlib.pyplot as plt
import numpy as np
from geoh5py.objects import BlockModel, Curve, Octree
from geoh5py.workspace import Workspace
from ipywidgets.widgets import HBox, VBox
from scipy.interpolate import LinearNDInterpolator
from scipy.spatial import cKDTree

from geoapps.plotting import plot_plan_data_selection
from geoapps.selection import ObjectDataSelection
from geoapps.inversion import TopographyOptions
from geoapps.utils import (
    export_curve_2_shapefile,
    export_grid_2_geotiff,
    object_2_dataframe,
    octree_2_treemesh,
)


def export_widget(h5file):
    """
    General widget to export geoh5 objects to different file formats.
    Currently supported:
    shapefiles: *.shp
    """
    workspace = Workspace(h5file)

    dsep = os.path.sep
    out_dir = (
        dsep.join(os.path.dirname(os.path.abspath(h5file)).split(dsep)) + os.path.sep
    )

    def save_selection(_):
        if export.value:

            export.value = False  # Reset button

            entity = workspace.get_entity(selection.objects.value)[0]

            if selection.data.value:

                data_values = {}

                for key in selection.data.value:
                    if entity.get_data(key):
                        data_values[key] = entity.get_data(key)[0].values.copy()
                        data_values[key][
                            (data_values[key] > 1e-38) * (data_values[key] < 2e-38)
                        ] = no_data_value.value
            else:
                data_values = {}

            if file_type.value == "csv":
                dataframe = object_2_dataframe(entity, fields=list(data_values.keys()))
                dataframe.to_csv(f"{out_dir + export_as.value}" + ".csv", index=False)

            elif file_type.value == "ESRI shapefile":

                assert isinstance(
                    entity, Curve
                ), f"Only Curve objects are support for type {file_type.value}"

                if selection.data.value:
                    for key in selection.data.value:
                        out_name = re.sub(
                            "[^0-9a-zA-Z]+", "_", export_as.value + "_" + key
                        )
                        export_curve_2_shapefile(
                            entity,
                            attribute=key,
                            file_name=out_dir + out_name,
                            epsg=epsg_code.value,
                        )
                        print(f"Object saved to {out_dir + out_name + '.shp'}")
                else:
                    out_name = re.sub("[^0-9a-zA-Z]+", "_", export_as.value)
                    export_curve_2_shapefile(
                        entity, file_name=out_dir + out_name, epsg=epsg_code.value,
                    )
                    print(f"Object saved to {out_dir + out_name + '.shp'}")

            elif file_type.value == "geotiff":

                for key in selection.data.value:
                    name = out_dir + export_as.value + "_" + key + ".tif"
                    if entity.get_data(key):

                        export_grid_2_geotiff(
                            entity.get_data(key)[0],
                            name,
                            epsg_code.value,
                            data_type=data_type.value,
                        )

                        if data_type.value == "RGB":
                            fig, ax = plt.figure(), plt.subplot()
                            plt.gca().set_visible(False)
                            (ax, im), _, _, _ = plot_plan_data_selection(
                                entity, entity.get_data(key)[0], ax=ax
                            )
                            plt.colorbar(im, fraction=0.02)
                            plt.savefig(
                                out_dir + export_as.value + "_" + key + "_Colorbar.png",
                                dpi=300,
                                bbox_inches="tight",
                            )

                        print(f"Object saved to {name}")

            elif file_type.value == "UBC format":

                assert isinstance(
                    entity, (Octree, BlockModel)
                ), "Export available for BlockModel or Octree only"
                if isinstance(entity, Octree):
                    mesh = octree_2_treemesh(entity)

                    models = {}
                    for key, item in data_values.items():
                        ind = np.argsort(mesh._ubc_order)

                        data_obj = entity.get_data(key)[0]
                        models[out_dir + data_obj.name + "_" + key + ".mod"] = item[ind]
                    mesh.writeUBC(out_dir + export_as.value + ".msh", models=models)

                else:

                    mesh = discretize.TensorMesh(
                        [
                            np.abs(entity.u_cells),
                            np.abs(entity.v_cells),
                            np.abs(entity.z_cells[::-1]),
                        ]
                    )

                    # Move the origin to the bottom SW corner
                    mesh.x0 = [
                        entity.origin["x"] + entity.u_cells[entity.u_cells < 0].sum(),
                        entity.origin["y"] + entity.v_cells[entity.v_cells < 0].sum(),
                        entity.origin["z"] + entity.z_cells[entity.z_cells < 0].sum(),
                    ]

                    mesh.writeUBC(out_dir + export_as.value + ".msh")

                    if any(data_values):
                        for key, item in data_values.items():

                            if mesh.x0[2] == entity.origin["z"]:
                                values = item.copy()
                                values = values.reshape(
                                    (mesh.nCz, mesh.nCx, mesh.nCy), order="F"
                                )[::-1, :, :]
                                values = values.reshape((-1, 1), order="F")
                            else:
                                values = item

                            np.savetxt(out_dir + key + ".mod", values)

    def update_options(_):

        if file_type.value in ["ESRI shapefile"]:
            type_widget.children = [file_type, epsg_code]
        elif file_type.value in ["geotiff"]:
            type_widget.children = [file_type, VBox([epsg_code, data_type])]
        else:
            type_widget.children = [file_type]

    file_type = widgets.Dropdown(
        options=["ESRI shapefile", "csv", "geotiff", "UBC format"],
        value="csv",
        description="Export type",
    )

    data_type = widgets.RadioButtons(options=["float", "RGB",], description="Type:")
    no_data_value = widgets.FloatText(description="no-data-value", value=-99999,)

    epsg_code = widgets.Text(description="EPSG code:", indent=False, disabled=False)

    type_widget = HBox([file_type])

    file_type.observe(update_options)

    selection = ObjectDataSelection(h5file=h5file, select_multiple=True)

    def update_name(_):
        export_as.value = selection.objects.value.replace(":", "_")

    selection.objects.observe(update_name, names="value")
    export = widgets.ToggleButton(
        value=False,
        description="Export",
        button_style="danger",
        tooltip="Description",
        icon="check",
    )

    export.observe(save_selection, names="value")

    export_as = widgets.Text(
        value=selection.objects.value,
        description="Save as:",
        indent=False,
        disabled=False,
    )
    return VBox(
        [HBox([selection.widget, no_data_value]), type_widget, export_as, export]
    )


def object_to_object_interpolation(h5file):
    def out_update(_):

        if out_mode.value == "To Object:":

            out_panel.children = [out_mode, mesh_dropdown]

        else:
            out_panel.children = [out_mode, new_grid_panel]

    def interpolate_call(_):

        if interpolate.value:

            object_from = workspace.get_entity(selection.objects.value)[0]

            if hasattr(object_from, "centroids"):
                xyz = object_from.centroids.copy()
            elif hasattr(object_from, "vertices"):
                xyz = object_from.vertices.copy()

            # Create a tree for the input mesh
            tree = cKDTree(xyz)

            if out_mode.value == "To Object:":

                object_to = workspace.get_entity(mesh_dropdown.value)[0]

                if hasattr(object_to, "centroids"):
                    xyz_out = object_to.centroids.copy()
                elif hasattr(object_to, "vertices"):
                    xyz_out = object_to.vertices.copy()

            else:

                ref_in = workspace.get_entity(ref_dropdown.value)[0]

                if hasattr(ref_in, "centroids"):
                    xyz_ref = ref_in.centroids
                elif hasattr(ref_in, "vertices"):
                    xyz_ref = ref_in.vertices

                # Find extent of grid
                h = np.asarray(core_cell_size.value.split(",")).astype(float).tolist()

                pads = (
                    np.asarray(padding_distance.value.split(",")).astype(float).tolist()
                )

                # Use discretize to build a tensor mesh
                mesh = discretize.utils.meshutils.mesh_builder_xyz(
                    xyz_ref,
                    h,
                    padding_distance=[
                        [pads[0], pads[1]],
                        [pads[2], pads[3]],
                        [pads[4], pads[5]],
                    ],
                    depth_core=depth_core.value,
                    expansion_factor=expansion_fact.value,
                )

                object_to = BlockModel.create(
                    workspace,
                    origin=[mesh.x0[0], mesh.x0[1], xyz_ref[:, 2].max()],
                    u_cell_delimiters=mesh.vectorNx - mesh.x0[0],
                    v_cell_delimiters=mesh.vectorNy - mesh.x0[1],
                    z_cell_delimiters=-(xyz_ref[:, 2].max() - mesh.vectorNz[::-1]),
                    name=new_grid.value,
                )

                # Try to recenter on nearest
                # Find nearest cells
                rad, ind = tree.query(object_to.centroids)
                ind_nn = np.argmin(rad)

                d_xyz = object_to.centroids[ind_nn, :] - xyz[ind[ind_nn], :]

                object_to.origin = np.r_[object_to.origin.tolist()] - d_xyz

                xyz_out = object_to.centroids.copy()

            values = {}
            for field in selection.data.value:
                model_in = object_from.get_data(field)[0]
                values[field] = model_in.values.copy()

                values[field][values[field] == no_data_value.value] = np.nan
                if space.value == "Log":
                    values[field] = np.log(values[field])

            values_interp = {}
            if method.value == "Linear":

                for key, value in values.items():
                    F = LinearNDInterpolator(xyz, value)
                    values_interp[key] = F(xyz_out)

            elif method.value == "Inverse Distance":

                angle = np.deg2rad((450.0 - np.asarray(skew_angle.value)) % 360.0)

                rotation = np.r_[
                    np.c_[np.cos(angle), np.sin(angle)],
                    np.c_[-np.sin(angle), np.cos(angle)],
                ]

                center = np.mean(xyz, axis=0).reshape((3, 1))
                xyz -= np.kron(center, np.ones(xyz.shape[0])).T
                xyz[:, :2] = np.dot(rotation, xyz[:, :2].T).T
                xyz[:, 1] *= skew_factor.value

                tree = cKDTree(xyz)

                xyz_out -= np.kron(center, np.ones(xyz_out.shape[0])).T
                xyz_out[:, :2] = np.dot(rotation, xyz_out[:, :2].T).T
                xyz_out[:, 1] *= skew_factor.value

                # Find nearest cells
                rad, ind = tree.query(xyz_out, 8)

                for key, value in values.items():
                    values_interp[key] = np.zeros(xyz_out.shape[0])
                    weight = np.zeros(xyz_out.shape[0])

                    for ii in range(8):
                        values_interp[key] += value[ind[:, ii]] / (rad[:, ii] + 1e-1)
                        weight += 1.0 / (rad[:, ii] + 1e-1)

                    values_interp[key] /= weight

            else:
                # Find nearest cells
                for key, value in values.items():
                    rad, ind = tree.query(xyz_out)
                    values_interp[key] = value[ind]

            for key in values_interp.keys():
                if space.value == "Log":
                    values_interp[key] = np.exp(values_interp[key])

                values_interp[key][np.isnan(values_interp[key])] = no_data_value.value

                if method.value == "Inverse Distance":
                    values_interp[key][
                        rad[:, 0] > max_distance.value
                    ] = no_data_value.value
                    if max_depth.value is not None:
                        values_interp[key][
                            np.abs(xyz_out[:, 2] - xyz[ind[:, 0], 2]) > max_depth.value
                        ] = no_data_value.value

                else:
                    values_interp[key][rad > max_distance.value] = no_data_value.value

                    if max_depth.value is not None:
                        values_interp[key][
                            np.abs(xyz_out[:, 2] - xyz[ind, 2]) > max_depth.value
                        ] = no_data_value.value

            if topo_options.options_button.value == "Object":
                topo_obj = workspace.get_entity(topo_options.objects.value)[0]
                if getattr(topo_obj, "vertices", None) is not None:
                    topo = topo_obj.vertices
                else:
                    topo = topo_obj.centroids

                if topo_options.value.value != "Vertices":
                    topo[:, 2] = topo_obj.get_data(topo_options.value.value)[0].values

                xyz_out = object_to.centroids.copy()
                F = LinearNDInterpolator(topo[:, :2], topo[:, 2])
                z_interp = F(xyz_out[:, :2])

                ind_nan = np.isnan(z_interp)
                if any(ind_nan):
                    tree = cKDTree(topo[:, :2])
                    _, ind = tree.query(xyz_out[ind_nan, :2])
                    z_interp[ind_nan] = topo[ind, 2]

                for key in values_interp.keys():
                    values_interp[key][xyz_out[:, 2] > z_interp] = no_data_value.value
            elif (
                topo_options.options_button.value == "Constant"
                and topo_options.constant.value is not None
            ):
                xyz_out = object_to.centroids.copy()
                for key in values_interp.keys():
                    values_interp[key][
                        xyz_out[:, 2] > topo_options.constant.value
                    ] = no_data_value.value

            if xy_extent.value is not None:
                xy_ref = workspace.get_entity(xy_extent.value)[0]
                if hasattr(xy_ref, "centroids"):
                    xy_ref = xy_ref.centroids
                elif hasattr(xy_ref, "vertices"):
                    xy_ref = xy_ref.vertices

                tree = cKDTree(xy_ref[:, :2])
                rad, _ = tree.query(xyz_out[:, :2])
                for key in values_interp.keys():
                    values_interp[key][rad > max_distance.value] = no_data_value.value

            for key in values_interp.keys():
                object_to.add_data({key + "_interp": {"values": values_interp[key]}})

            interpolate.value = False
            workspace.finalize()

    workspace = Workspace(h5file)

    names = list(workspace.list_objects_name.values())

    selection = ObjectDataSelection(h5file=h5file, select_multiple=True)

    out_mode = widgets.RadioButtons(
        options=["To Object:", "Create 3D Grid"], value="To Object:", disabled=False
    )

    out_mode.observe(out_update)
    mesh_dropdown = widgets.Dropdown(options=names)

    ref_dropdown = widgets.Dropdown(options=names, description="XY Extent from:",)

    def object_pick(_):
        ref_dropdown.value = selection.objects.value

    selection.objects.observe(object_pick)

    new_grid = widgets.Text(
        value="InterpGrid", description="New grid name:", disabled=False,
    )

    core_cell_size = widgets.Text(
        value="25, 25, 25", description="Smallest cells", disabled=False,
    )

    depth_core = widgets.FloatText(
        value=500, description="Core depth (m)", disabled=False,
    )

    padding_distance = widgets.Text(
        value="0, 0, 0, 0, 0, 0",
        description="Pad Distance (W, E, N, S, D, U)",
        disabled=False,
    )

    expansion_fact = widgets.FloatText(
        value=1.05, description="Expansion factor", disabled=False,
    )

    space = widgets.RadioButtons(
        options=["Linear", "Log"], value="Linear", disabled=False
    )

    def method_update(_):

        if method.value == "Inverse Distance":
            method_panel.children = [method, method_skew]
        else:
            method_panel.children = [method]

    method = widgets.RadioButtons(
        options=["Nearest", "Linear", "Inverse Distance"],
        value="Nearest",
        disabled=False,
    )

    max_distance = widgets.FloatText(value=1e3, description="Maximum distance XY (m)",)

    max_depth = widgets.FloatText(value=1e3, description="Maximum distance Z (m)",)

    no_data_value = widgets.FloatText(value=-99999, description="No-Data-Value",)

    skew_angle = widgets.FloatText(
        value=0, description="Azimuth (d.dd)", disabled=False,
    )

    skew_factor = widgets.FloatText(value=1, description="Factor", disabled=False,)

    method_skew = VBox([widgets.Label("Skew interpolation"), skew_angle, skew_factor])
    method_panel = HBox([method])

    method.observe(method_update)
    interpolate = widgets.ToggleButton(
        value=False, description="Interpolate", icon="check"
    )

    interpolate.observe(interpolate_call)

    topo_options = TopographyOptions(h5file=h5file)
    topo_options.offset.disabled = True
    topo_options.options_button.options = ["Object", "Constant", "None"]
    topo_options.options_button.value = "Object"
    # topography, z_value = ObjectDataSelection(h5file)
    # z_value.options = list(z_value.options) + ["Vertices"]

    xy_extent = widgets.Dropdown(
        options=[None] + names, description="Trim xy extent with:",
    )

    new_grid_panel = VBox(
        [
            new_grid,
            ref_dropdown,
            core_cell_size,
            depth_core,
            padding_distance,
            expansion_fact,
        ]
    )

    out_panel = HBox([out_mode, mesh_dropdown])

    return VBox(
        [
            HBox(
                [
                    VBox([widgets.Label("Input"), selection.widget]),
                    VBox([widgets.Label("Output"), out_panel]),
                ]
            ),
            VBox(
                [
                    widgets.Label("Interpolation Parameters"),
                    HBox(
                        [
                            VBox([widgets.Label("Space"), space]),
                            VBox([widgets.Label("Method"), method_panel]),
                        ]
                    ),
                    no_data_value,
                    max_distance,
                    max_depth,
                    VBox([widgets.Label("Cut with topo"), topo_options.widget]),
                    xy_extent,
                ]
            ),
            interpolate,
        ]
    )



import properties
import numpy as np
import warnings

from . import Maps
from . import Utils


class SphinxProp:
    """
    Update the auto-documenter from properties
    https://github.com/3ptscience/properties/issues/153
    """

    def sphinx_class(self):
        return ":class:`{cls} <{ref}>`".format(
            cls=self.__class__.__name__, ref=f"SimPEG.Props.{self.__class__.__name__}",
        )


class Array(SphinxProp, properties.Array):

    class_info = "a numpy, Zero or Identity array"

    def validate(self, instance, value):
        if isinstance(value, (Utils.Zero, Utils.Identity)):
            return value
        return super().validate(instance, value)


class Float(SphinxProp, properties.Float):

    class_info = "a float, Zero or Identity"

    def validate(self, instance, value):
        if isinstance(value, (Utils.Zero, Utils.Identity)):
            return value
        return super().validate(instance, value)


class Integer(SphinxProp, properties.Integer):

    class_info = "an Integer or *"

    def validate(self, instance, value):
        if isinstance(value, str):
            assert value == "*", f"value must be an integer or *, not {value}"
            return value
        return super().validate(instance, value)


class Model(SphinxProp, properties.Array):

    class_info = "a numpy array"
    _required = False


class Mapping(SphinxProp, properties.Property):

    class_info = "a SimPEG Map"
    _required = False

    @property
    def prop(self):
        return getattr(self, "_prop", None)

    @prop.setter
    def prop(self, value):
        assert isinstance(value, PhysicalProperty)
        value._mapping = self  # Skip the setter
        self._prop = value

    @property
    def reciprocal(self):
        if self.prop and self.prop.reciprocal:
            return self.prop.reciprocal.mapping

    @property
    def reciprocal_prop(self):
        if self.prop and self.prop.reciprocal:
            return self.prop.reciprocal

    def clear_props(self, instance):
        for prop in (self.prop, self.reciprocal_prop, self.reciprocal):
            if prop is not None:
                if prop.name in instance._props:
                    delattr(instance, prop.name)
                else:
                    setattr(instance, prop.name, None)

    def validate(self, instance, value):
        if not isinstance(value, Maps.IdentityMap):
            self.error(instance, value)
        return value

    def get_property(self):

        scope = self

        def fget(self):
            value = self._get(scope.name)
            if value is not None:
                return value
            if scope.reciprocal is None:
                return None
            reciprocal = self._get(scope.reciprocal.name)
            if reciprocal is None:
                return None
            return Maps.ReciprocalMap() * reciprocal

        def fset(self, value):
            if value is not properties.utils.undefined:
                value = scope.validate(self, value)
            self._set(scope.name, value)
            if value is not properties.utils.undefined:
                scope.clear_props(self)

        def fdel(self):
            self._set(scope.name, properties.utils.undefined)

        return property(fget=fget, fset=fset, fdel=fdel, doc=scope.doc)

    def as_pickle(self, instance):
        return instance._get(self.name)


class PhysicalProperty(SphinxProp, properties.Property):

    class_info = "a physical property"
    reciprocal = None
    _required = False

    @property
    def mapping(self):
        return getattr(self, "_mapping", None)

    @mapping.setter
    def mapping(self, value):
        assert isinstance(value, Mapping)
        value._prop = self  # Skip the setter
        self._mapping = value

    def clear_mappings(self, instance):
        if self.mapping is not None:
            if self.mapping.name in instance._props:
                delattr(instance, self.mapping.name)
            else:
                setattr(instance, self.mapping.name, None)
        if self.reciprocal is not None:
            if self.reciprocal.mapping is not None:
                if self.reciprocal.mapping.name in instance._props:
                    delattr(instance, self.reciprocal.mapping.name)
                else:
                    setattr(instance, self.reciprocal.mapping.name, None)

    def validate(self, instance, value):
        assert isinstance(
            value, (np.ndarray, float)
        ), "Physical properties must be numpy arrays or floats."
        return value

    def get_property(self):

        scope = self

        def fget(self):

            default = self._get(scope.name)
            if default is not None:
                return default
            if scope.reciprocal:
                default = self._get(scope.reciprocal.name)
                if default is not None:
                    # set by default reciprocal
                    return 1.0 / default
            if scope.mapping is None and scope.reciprocal is None:
                return None
            if scope.mapping is None:
                # look to the reciprocal
                if scope.reciprocal.mapping is None:
                    # there is no reciprocal mapping
                    reciprocal_val = self._get(scope.reciprocal.name)
                    if reciprocal_val is None:
                        raise AttributeError(
                            "A default for {}/{} has not been set".format(
                                scope.name, scope.reciprocal.name
                            )
                        )
                # Set by mapped reciprocal
                return 1.0 / getattr(self, scope.reciprocal.name)

            mapping = getattr(self, scope.mapping.name)
            if mapping is None:
                raise AttributeError(
                    "A default `{}` or mapping `{}` has not been set.".format(
                        scope.name, scope.mapping.name
                    )
                )
            if self.model is None:
                raise AttributeError(
                    f"A `model` is required for physical property {scope.name}"
                )
            return mapping * self.model

        def fset(self, value):
            if value is not properties.utils.undefined:
                value = scope.validate(self, value)
                if scope.reciprocal:
                    delattr(self, scope.reciprocal.name)
            self._set(scope.name, value)
            if value is not properties.utils.undefined:
                scope.clear_mappings(self)

        def fdel(self):
            self._set(scope.name, properties.utils.undefined)

        return property(fget=fget, fset=fset, fdel=fdel, doc=scope.doc)

    def as_pickle(self, instance):
        return instance._get(self.name)

    def summary(self, instance):
        default = instance._get(self.name)
        if default is not None:
            return f"[*] {self.name}: set by default value"
        if self.reciprocal:
            default = instance._get(self.reciprocal.name)
            if default is not None:
                return "[*] {}: set by default reciprocal: 1.0 / {}".format(
                    self.name, self.reciprocal.name
                )
        if self.mapping is None and self.reciprocal is None:
            return f"[ ] {self.name}: property not set"
        if self.mapping is None:
            if self.reciprocal.mapping is None:
                # there is no reciprocal mapping
                reciprocal_val = instance._get(self.reciprocal.name)
                if reciprocal_val is None:
                    return "[ ] {}: default for {}/{} not set".format(
                        self.name, self.name, self.reciprocal.name
                    )
            return "[*] {}: set by mapped reciprocal 1.0 / ({} * {})".format(
                self.name, self.reciprocal.mapping.name, self.reciprocal.name
            )

        mapping = getattr(instance, self.mapping.name)

        if mapping is None:
            return "[ ] {}: default `{}` or mapping `{}` not set".format(
                self.name, self.name, self.mapping.name
            )
        if instance.model is None:
            return ("[ ] {}: model({}) required, from active `{}` mapping: {}").format(
                self.name, mapping.shape[1], self.mapping.name, str(mapping)
            )

        correct_shape = mapping.shape[1] == "*" or mapping.shape[1] == len(
            instance.model
        )

        if correct_shape:
            return "[*] {}: set by the `{}` mapping: {} * model({})".format(
                self.name, self.mapping.name, str(mapping), len(instance.model)
            )

        return "[ ] {}: incorrect mapping/model shape: {} * model({})".format(
            self.name, str(mapping), len(instance.model)
        )


class Derivative(SphinxProp, properties.GettableProperty):

    physical_property = None

    @property
    def mapping(self):
        """The mapping looks through to the physical property map."""
        if self.physical_property is None:
            return None
        return self.physical_property.mapping

    def get_property(self):

        scope = self

        def fget(self):
            if scope.physical_property is None:
                return Utils.Zero()
            if scope.mapping is None:
                return Utils.Zero()
            mapping = getattr(self, scope.mapping.name)
            if mapping is None:
                return Utils.Zero()
            if self.model is None:
                return Utils.Zero()

            return mapping.deriv(self.model)

        return property(fget=fget, doc=scope.doc)


def Invertible(help, default=None):

    mapping = Mapping(f"Mapping of {help} to the inversion model.")

    physical_property = PhysicalProperty(help, mapping=mapping)
    if default is not None:
        physical_property.default = default

    property_derivative = Derivative(
        f"Derivative of {help} wrt the model.", physical_property=physical_property,
    )

    return physical_property, mapping, property_derivative


def Reciprocal(prop1, prop2):
    prop1.reciprocal = prop2
    prop2.reciprocal = prop1


class BaseSimPEG(properties.HasProperties):
    """"""


class HasModel(BaseSimPEG):

    model = Model("Inversion model.")

    @property
    def _all_map_names(self):
        """Returns all Mapping properties"""
        return sorted([k for k in self._props if isinstance(self._props[k], Mapping)])

    @property
    def _act_map_names(self):
        """Returns all active Mapping properties"""
        return sorted([k for k in self._all_map_names if getattr(self, k) is not None])

    @property
    def needs_model(self):
        """True if a model is necessary"""
        return len(self._act_map_names) > 0

    @property
    def _has_nested_models(self):
        for k in self._props:
            if isinstance(self._props[k], properties.Instance) and issubclass(
                self._props[k].instance_class, HasModel
            ):
                return True
        return False

    @properties.validator("model")
    def _check_model_valid(self, change):
        """Checks the model length and necessity"""
        if change["value"] is properties.utils.undefined:
            return True

        if not self.needs_model and not self._has_nested_models:
            warnings.warn(
                "Cannot add model as there are no active mappings"
                ", choose from: ['{}']".format("', '".join(self._all_map_names))
            )
            return

        errors = []

        for name in self._act_map_names:
            mapping = getattr(self, name)
            correct_shape = mapping.shape[1] == "*" or mapping.shape[1] == len(
                change["value"]
            )
            if not correct_shape:
                errors += [
                    "{}: expected model of len({}) for {}".format(
                        name, mapping.shape[1], str(mapping)
                    )
                ]
        if len(errors) == 0:
            return True

        warnings.warn(
            "Model of len({}) incorrect shape for mappings: \n    {}".format(
                len(change["value"]), "\n    ".join(errors)
            )
        )

    @properties.validator
    def _check_valid(self):
        errors = []

        # Check if the model is necessary
        if self.needs_model and self.model is None:
            errors += ["model must not be None"]
        if not self.needs_model and self.model is not None:
            errors += ["there are no active maps, but a model is provided"]

        # Check each map is the same size
        shapes = []
        for name in self._act_map_names:
            shape = getattr(self, name).shape[1]
            if shape == "*":
                continue
            shapes += [shape]
        if not all(x == shapes[0] for x in shapes):
            errors += ["the mappings are not the same shape"]

        # Check that the model is the same size as the mappings
        if len(shapes) > 0 and self.model is not None:
            if not len(self.model) == shapes[0]:
                errors += ["the model must be len({})".format(shapes[0])]

        # Check each physical property
        check_boxes = sorted(
            [
                self._props[k].summary(self)
                for k in self._props
                if isinstance(self._props[k], PhysicalProperty)
            ]
        )
        for line in check_boxes:
            if line[:3] == "[ ]":
                errors += [line[4:]]

        if len(errors) == 0:
            return True

        raise ValueError(
            "The {} instance has the following errors: \n - {}".format(
                self.__class__.__name__, "\n - ".join(errors)
            )
        )

    def summary(self):
        prop_names = sorted(
            [k for k in self._props if isinstance(self._props[k], PhysicalProperty)]
        )

        out = ["Physical Properties:"]

        # Grab the physical property summaries
        for prop in prop_names:
            out += [" " + self._props[prop].summary(self)]

        # Grab the validation errors
        try:
            self.validate()
            out += ["", "All checks pass!"]
        except ValueError as e:
            out += ["", str(e)]

        return "\n".join(out)



import os
import numpy as np
import time
import matplotlib.pyplot as plt
from geoh5py.workspace import Workspace
from geoh5py.objects import Points
from ipywidgets import (
    Dropdown,
    ColorPicker,
    SelectMultiple,
    Text,
    IntSlider,
    Checkbox,
    FloatSlider,
    FloatText,
    VBox,
    HBox,
    ToggleButton,
    ToggleButtons,
    interactive_output,
    FloatLogSlider,
    Label,
    Layout,
    RadioButtons,
)
from geoapps.base import BaseApplication
from geoapps.utils import find_value, geophysical_systems, signal_processing_1d
from geoapps.selection import ObjectDataSelection, LineOptions


class EMLineProfiler(BaseApplication):
    groups = {
        "early": {
            "color": "blue",
            "channels": [],
            "gates": [],
            "defaults": ["early"],
            "inflx_up": [],
            "inflx_dwn": [],
            "peaks": [],
            "times": [],
            "values": [],
            "locations": [],
        },
        "early + middle": {
            "color": "cyan",
            "channels": [],
            "gates": [],
            "defaults": ["early", "middle"],
            "inflx_up": [],
            "inflx_dwn": [],
            "peaks": [],
            "times": [],
            "values": [],
            "locations": [],
        },
        "middle": {
            "color": "green",
            "channels": [],
            "gates": [],
            "defaults": ["middle"],
            "inflx_up": [],
            "inflx_dwn": [],
            "peaks": [],
            "times": [],
            "values": [],
            "locations": [],
        },
        "early + middle + late": {
            "color": "orange",
            "channels": [],
            "gates": [],
            "defaults": ["early", "middle", "late"],
            "inflx_up": [],
            "inflx_dwn": [],
            "peaks": [],
            "times": [],
            "values": [],
            "locations": [],
        },
        "middle + late": {
            "color": "yellow",
            "channels": [],
            "gates": [],
            "defaults": ["middle", "late"],
            "inflx_up": [],
            "inflx_dwn": [],
            "peaks": [],
            "times": [],
            "values": [],
            "locations": [],
        },
        "late": {
            "color": "red",
            "channels": [],
            "gates": [],
            "defaults": ["late"],
            "inflx_up": [],
            "inflx_dwn": [],
            "peaks": [],
            "times": [],
            "values": [],
            "locations": [],
        },
    }

    viz_param = '<IParameterList Version="1.0">\n <Colour>4278190335</Colour>\n <Transparency>0</Transparency>\n <Nodesize>9</Nodesize>\n <Nodesymbol>Sphere</Nodesymbol>\n <Scalenodesbydata>false</Scalenodesbydata>\n <Data></Data>\n <Scale>1</Scale>\n <Scalebyabsolutevalue>[NO STRING REPRESENTATION]</Scalebyabsolutevalue>\n <Orientation toggled="true">{\n    "DataGroup": "AzmDip",\n    "ManualWidth": true,\n    "Scale": false,\n    "ScaleLog": false,\n    "Size": 30,\n    "Symbol": "Tablet",\n    "Width": 30\n}\n</Orientation>\n</IParameterList>\n'

    def __init__(self, h5file):

        self.workspace = Workspace(h5file)
        self.em_system_specs = geophysical_systems.parameters()
        self.system = Dropdown(
            options=[
                key
                for key, specs in self.em_system_specs.items()
                if specs["type"] == "time"
            ],
            description="Time-Domain System:",
        )

        self._groups = self.groups
        self.group_list = SelectMultiple(description="")

        self.early = np.arange(8, 17).tolist()
        self.middle = np.arange(17, 28).tolist()
        self.late = np.arange(28, 40).tolist()

        self.data_selection = ObjectDataSelection(h5file=h5file, select_multiple=True)
        self.data_selection.objects.description = "Survey"

        self.model_selection = ObjectDataSelection(h5file=h5file)
        self.model_selection.objects.description = "1D Object:"
        self.model_selection.data.description = "Model"

        _, self.model_line_field = ObjectDataSelection(
            h5file=h5file, objects=self.model_selection.objects, find_value=["line"]
        )
        self.model_line_field.description = "Line field: "

        self.marker = {"left": "<", "right": ">"}

        def update_model_line_fields(_):
            self.model_line_field.options = self.model_field.options
            self.model_line_field.value = find_value(
                self.model_line_field.options, ["line"]
            )

        self.model_selection.data.observe(update_model_line_fields, names="options")

        def get_survey(_):
            if self.workspace.get_entity(self.data_objects.value):
                self.survey = self.workspace.get_entity(self.data_objects.value)[0]
                self.data_field.options = (
                    [p_g.name for p_g in self.survey.property_groups]
                    + ["^-- Groups --^"]
                    + list(self.data_field.options)
                )

        self.data_objects.observe(get_survey, names="value")
        self.data_objects.value = self.data_objects.options[0]
        get_survey("")

        self.lines = LineOptions(h5file, self.data_objects, select_multiple=False)
        self.lines.value.description = "Line"

        self.channels = SelectMultiple(description="Channels")
        self.group_default_early = Text(description="Early", value="9-16")
        self.group_default_middle = Text(description="Middle", value="17-27")
        self.group_default_late = Text(description="Late", value="28-40")

        def reset_default_bounds(_):
            self.reset_default_bounds()

        self.data = {}
        self.data_channel_options = {}

        def get_data(_):
            data = []

            groups = [p_g.name for p_g in self.survey.property_groups]
            channels = list(self.data_field.value)

            for channel in self.data_field.value:
                if channel in groups:
                    for prop in self.survey.get_property_group(channel).properties:
                        name = self.workspace.get_entity(prop)[0].name
                        if prop not in channels:
                            channels.append(name)

            self.channels.options = channels
            for channel in channels:
                if self.survey.get_data(channel):
                    self.data[channel] = self.survey.get_data(channel)[0]

            # Generate default groups
            self.reset_default_bounds()

            for key, widget in self.data_channel_options.items():
                widget.children[0].options = channels
                widget.children[0].value = find_value(channels, [key])

        self.data_field.observe(get_data, names="value")
        get_data("")

        def get_surf_model(_):
            if self.workspace.get_entity(self.model_objects.value):
                self.surf_model = self.workspace.get_entity(self.model_objects.value)[0]

        self.model_objects.observe(get_surf_model, names="value")

        def get_model(_):
            if self.surf_model.get_data(self.model_field.value):
                self.surf_model = self.surf_model.get_data(self.model_field.value)[0]

        self.model_objects.observe(get_model, names="value")

        self.smoothing = IntSlider(
            min=0, max=64, value=0, description="Running mean", continuous_update=False,
        )

        self.residual = Checkbox(description="Use residual", value=False)

        self.threshold = FloatSlider(
            value=50,
            min=10,
            max=90,
            step=5,
            continuous_update=False,
            description="Decay threshold (%)",
        )

        self.center = FloatSlider(
            value=0.5,
            min=0,
            max=1.0,
            step=0.001,
            description="Center (%)",
            disabled=False,
            continuous_update=False,
            orientation="horizontal",
        )

        self.auto_picker = ToggleButton(description="Pick target", value=True)

        self.focus = FloatSlider(
            value=1.0,
            min=0.025,
            max=1.0,
            step=0.005,
            description="Width (%)",
            disabled=False,
            continuous_update=False,
            orientation="horizontal",
        )

        self.zoom = VBox([self.center, self.focus])

        self.scale_button = RadioButtons(
            options=["linear", "symlog",], description="Vertical scaling",
        )

        def scale_update(_):
            if self.scale_button.value == "symlog":
                scale_panel.children = [
                    self.scale_button,
                    VBox([Label("Linear threshold"), self.scale_value]),
                ]
            else:
                scale_panel.children = [self.scale_button]

        self.scale_button.observe(scale_update)
        self.scale_value = FloatLogSlider(
            min=-18,
            max=10,
            step=0.5,
            base=10,
            value=1e-2,
            description="",
            continuous_update=False,
        )
        scale_panel = HBox([self.scale_button])

        def add_group(_):
            self.add_group()

        def channel_setter(caller):

            channel = caller["owner"]
            data_widget = self.data_channel_options[channel.header]
            data_widget.children[0].value = find_value(
                data_widget.children[0].options, [channel.header]
            )

        self.channel_selection = Dropdown(
            description="Time Gate",
            options=self.em_system_specs[self.system.value]["channels"].keys(),
        )

        def system_observer(_):

            system_specs = {}
            for key, time_gate in self.em_system_specs[self.system.value][
                "channels"
            ].items():
                system_specs[key] = f"{time_gate:.5e}"

            self.channel_selection.options = self.em_system_specs[self.system.value][
                "channels"
            ].keys()

            self.data_channel_options = {}
            for ind, (key, value) in enumerate(system_specs.items()):
                channel_selection = Dropdown(
                    description="Channel",
                    style={"description_width": "initial"},
                    options=self.channels.options,
                    value=find_value(self.channels.options, [key]),
                )
                channel_selection.header = key
                channel_selection.observe(channel_setter, names="value")

                channel_time = FloatText(description="Time (s)", value=value)

                self.data_channel_options[key] = VBox([channel_selection, channel_time])

        self.system.observe(system_observer)
        system_observer("")

        def channel_panel_update(_):
            self.channel_panel.children = [
                self.channel_selection,
                self.data_channel_options[self.channel_selection.value],
            ]

        self.channel_selection.observe(channel_panel_update, names="value")
        self.channel_panel = VBox(
            [
                self.channel_selection,
                self.data_channel_options[self.channel_selection.value],
            ]
        )

        self.group_default_early.observe(reset_default_bounds)
        self.group_default_middle.observe(reset_default_bounds)
        self.group_default_late.observe(reset_default_bounds)

        self.group_add = ToggleButton(description="<< Add New Group <<")
        self.group_name = Text(description="Name")
        self.group_color = ColorPicker(
            concise=False, description="Color", value="blue", disabled=False
        )
        self.group_add.observe(add_group)

        def highlight_selection(_):

            self.highlight_selection()

        self.group_list.observe(highlight_selection, names="value")
        self.markers = ToggleButton(description="Show all markers")

        def export_click(_):
            self.export_click()

        self.export = ToggleButton(description="Export", button_style="success")
        self.export.observe(export_click)

        def plot_data_selection(
            ind,
            smoothing,
            residual,
            markers,
            scale,
            scale_value,
            center,
            focus,
            groups,
            pick_trigger,
            x_label,
            threshold,
        ):
            self.plot_data_selection(
                ind,
                smoothing,
                residual,
                markers,
                scale,
                scale_value,
                center,
                focus,
                groups,
                pick_trigger,
                x_label,
                threshold,
            )

        def plot_model_selection(ind, center, focus):
            self.plot_model_selection(ind, center, focus)

        self.x_label = ToggleButtons(
            options=["Distance", "Easting", "Northing"],
            value="Distance",
            description="X-axis label:",
        )
        plotting = interactive_output(
            plot_data_selection,
            {
                "ind": self.lines.lines,
                "smoothing": self.smoothing,
                "residual": self.residual,
                "markers": self.markers,
                "scale": self.scale_button,
                "scale_value": self.scale_value,
                "center": self.center,
                "focus": self.focus,
                "groups": self.group_list,
                "pick_trigger": self.auto_picker,
                "x_label": self.x_label,
                "threshold": self.threshold,
            },
        )

        section = interactive_output(
            plot_model_selection,
            {"ind": self.lines.lines, "center": self.center, "focus": self.focus,},
        )

        self.model_panel = VBox(
            [
                HBox(
                    [
                        self.model_objects,
                        VBox([self.model_field, self.model_line_field]),
                    ]
                ),
                section,
            ]
        )

        self.show_model = Checkbox(description="Show model", value=False)

        def show_model_trigger(_):
            self.show_model_trigger()

        self.show_model.observe(show_model_trigger)

        self.data_panel = VBox(
            [
                HBox(
                    [
                        VBox([self.data_objects, self.data_field]),
                        VBox([self.system, self.channel_panel]),
                    ]
                ),
                self.lines.widget,
                plotting,
                self.x_label,
                HBox(
                    [
                        VBox(
                            [
                                HBox(
                                    [
                                        VBox(
                                            [self.zoom, scale_panel],
                                            layout=Layout(width="50%"),
                                        ),
                                        VBox(
                                            [
                                                self.smoothing,
                                                self.residual,
                                                self.threshold,
                                            ],
                                            layout=Layout(width="50%"),
                                        ),
                                    ]
                                ),
                                HBox([self.markers, self.auto_picker]),
                                VBox(
                                    [
                                        Label("Groups"),
                                        HBox(
                                            [
                                                Label("Defaults"),
                                                self.group_default_early,
                                                self.group_default_middle,
                                                self.group_default_late,
                                            ]
                                        ),
                                        HBox(
                                            [
                                                self.group_list,
                                                VBox(
                                                    [
                                                        self.channels,
                                                        self.group_name,
                                                        self.group_color,
                                                        self.group_add,
                                                    ]
                                                ),
                                            ]
                                        ),
                                        HBox([self.export, self.live_link_widget,]),
                                    ]
                                ),
                            ]
                        )
                    ]
                ),
            ]
        )

        self._widget = VBox([self.data_panel, self.show_model])

    def set_default_groups(self, channels):
        """
        Assign TEM channel for given gate #
        """
        # Reset channels
        for group in self.groups.values():
            if len(group["defaults"]) > 0:
                group["channels"] = []
                group["inflx_up"] = []
                group["inflx_dwn"] = []
                group["peaks"] = []
                group["mad_tau"] = []
                group["times"] = []
                group["values"] = []
                group["locations"] = []

        for ind, channel in enumerate(channels):
            for group in self.groups.values():
                if ind in group["gates"]:
                    group["channels"].append(channel)

        self.group_list.options = self.groups.keys()
        self.group_list.value = []

    def add_group(self):
        """
        Add a group to the list of groups
        """
        if self.group_add.value:

            if self.group_name.value not in self.group_list.options:
                self.group_list.options = list(self.group_list.options) + [
                    self.group_name.value
                ]

            self.groups[self.group_name.value] = {
                "color": self.group_color.value,
                "channels": list(self.channels.value),
                "inflx_up": [],
                "inflx_dwn": [],
                "peaks": [],
                "mad_tau": [],
                "times": [],
                "values": [],
                "locations": [],
            }
            self.group_add.value = False

    def export_click(self):
        if self.export.value:
            for group in self.group_list.value:

                for (
                    ind,
                    (channel, locations, peaks, inflx_dwn, inflx_up, vals, times),
                ) in enumerate(
                    zip(
                        self.groups[group]["channels"],
                        self.groups[group]["locations"],
                        self.groups[group]["peaks"],
                        self.groups[group]["inflx_dwn"],
                        self.groups[group]["inflx_up"],
                        self.groups[group]["values"],
                        self.groups[group]["times"],
                    )
                ):

                    if ind == 0:
                        cox_x = self.lines.profile.interp_x(peaks[0])
                        cox_y = self.lines.profile.interp_y(peaks[0])
                        cox_z = self.lines.profile.interp_z(peaks[0])
                        cox = np.r_[cox_x, cox_y, cox_z]

                        # Compute average dip
                        left_ratio = np.abs(
                            (peaks[1] - inflx_up[1]) / (peaks[0] - inflx_up[0])
                        )
                        right_ratio = np.abs(
                            (peaks[1] - inflx_dwn[1]) / (peaks[0] - inflx_dwn[0])
                        )

                        if left_ratio > right_ratio:
                            ratio = right_ratio / left_ratio
                            azm = (
                                450.0
                                - np.rad2deg(
                                    np.arctan2(
                                        (
                                            self.lines.profile.interp_y(inflx_up[0])
                                            - cox_y
                                        ),
                                        (
                                            self.lines.profile.interp_x(inflx_up[0])
                                            - cox_x
                                        ),
                                    )
                                )
                            ) % 360.0
                        else:
                            ratio = left_ratio / right_ratio
                            azm = (
                                450.0
                                - np.rad2deg(
                                    np.arctan2(
                                        (
                                            self.lines.profile.interp_y(inflx_dwn[0])
                                            - cox_y
                                        ),
                                        (
                                            self.lines.profile.interp_x(inflx_dwn[0])
                                            - cox_x
                                        ),
                                    )
                                )
                            ) % 360.0

                        dip = np.rad2deg(np.arcsin(ratio))
                    tau = self.groups[group]["mad_tau"]

                if self.workspace.get_entity(group):
                    points = self.workspace.get_entity(group)[0]
                    azm_data = points.get_data("azimuth")[0]
                    azm_vals = azm_data.values.copy()
                    dip_data = points.get_data("dip")[0]
                    dip_vals = dip_data.values.copy()

                    tau_data = points.get_data("tau")[0]
                    tau_vals = tau_data.values.copy()

                    points.vertices = np.vstack([points.vertices, cox.reshape((1, 3))])
                    azm_data.values = np.hstack([azm_vals, azm])
                    dip_data.values = np.hstack([dip_vals, dip])
                    tau_data.values = np.hstack([tau_vals, tau])

                else:
                    # if self.workspace.get_entity(group)
                    # parent =
                    points = Points.create(
                        self.workspace, name=group, vertices=cox.reshape((1, 3))
                    )
                    points.add_data(
                        {
                            "azimuth": {"values": np.asarray(azm)},
                            "dip": {"values": np.asarray(dip)},
                            "tau": {"values": np.asarray(tau)},
                            # "Visual Parameters": {"values": self.viz_param}
                        }
                    )
                    group = points.find_or_create_property_group(
                        name="AzmDip", property_group_type="Dip direction & dip"
                    )
                    group.properties = [
                        points.get_data("azimuth")[0].uid,
                        points.get_data("dip")[0].uid,
                    ]

                if self.live_link.value:
                    if not os.path.exists(self.live_link_path.value):
                        os.mkdir(self.live_link_path.value)

                    temp_geoh5 = os.path.join(
                        self.live_link_path.value, f"temp{time.time():.3f}.geoh5"
                    )
                    ws_out = Workspace(temp_geoh5)
                    points.copy(parent=ws_out)

            self.export.value = False
            self.workspace.finalize()

    def highlight_selection(self):
        """
        Highlight the group choice
        """
        highlights = []
        for group in self.group_list.value:
            highlights += self.groups[group]["channels"]
            self.group_color.value = self.groups[group]["color"]
        self.channels.value = highlights

    def plot_data_selection(
        self,
        ind,
        smoothing,
        residual,
        markers,
        scale,
        scale_value,
        center,
        focus,
        groups,
        pick_trigger,
        x_label,
        threshold,
    ):

        fig = plt.figure(figsize=(12, 8))
        ax2 = plt.subplot()

        self.line_update()

        for group in self.groups.values():
            group["inflx_up"] = []
            group["inflx_dwn"] = []
            group["peaks"] = []
            group["mad_tau"]
            group["times"] = []
            group["values"] = []
            group["locations"] = []

        if (
            getattr(self, "survey", None) is None
            or getattr(self.lines, "profile", None) is None
        ):
            return

        center_x = center * self.lines.profile.locations_resampled[-1]

        if residual != self.lines.profile.residual:
            self.lines.profile.residual = residual
            self.line_update()

        if smoothing != self.lines.profile.smoothing:
            self.lines.profile.smoothing = smoothing
            self.line_update()

        lims = np.searchsorted(
            self.lines.profile.locations_resampled,
            [
                (center - focus / 2.0) * self.lines.profile.locations_resampled[-1],
                (center + focus / 2.0) * self.lines.profile.locations_resampled[-1],
            ],
        )

        sub_ind = np.arange(lims[0], lims[1])

        channels = []
        for group in self.group_list.value:
            channels += self.groups[group]["channels"]

        if len(channels) == 0:
            channels = self.channels.options

        times = {}
        for channel in self.data_channel_options.values():
            times[channel.children[0].value] = channel.children[1].value

        y_min, y_max = np.inf, -np.inf
        for channel, d in self.data.items():

            if channel not in times.keys():
                continue

            if channel not in channels:
                continue

            self.lines.profile.values = d.values[self.survey.line_indices].copy()
            locs, values = (
                self.lines.profile.locations_resampled[sub_ind],
                self.lines.profile.values_resampled[sub_ind],
            )

            y_min = np.min([values.min(), y_min])
            y_max = np.max([values.max(), y_max])

            ax2.plot(locs, values, color=[0.5, 0.5, 0.5, 1])

            if not residual:
                raw = self.lines.profile._values_resampled_raw[sub_ind]
                ax2.fill_between(
                    locs, values, raw, where=raw > values, color=[1, 0, 0, 0.5]
                )
                ax2.fill_between(
                    locs, values, raw, where=raw < values, color=[0, 0, 1, 0.5]
                )

            dx = self.lines.profile.derivative(order=1)[sub_ind]
            ddx = self.lines.profile.derivative(order=2)[sub_ind]

            peaks = np.where((np.diff(np.sign(dx)) != 0) * (ddx[1:] < 0))[0]
            lows = np.where((np.diff(np.sign(dx)) != 0) * (ddx[1:] > 0))[0]

            up_inflx = np.where((np.diff(np.sign(ddx)) != 0) * (dx[1:] > 0))[0]
            dwn_inflx = np.where((np.diff(np.sign(ddx)) != 0) * (dx[1:] < 0))[0]

            if markers:
                ax2.scatter(locs[peaks], values[peaks], s=100, color="r", marker="v")
                ax2.scatter(locs[lows], values[lows], s=100, color="b", marker="^")
                ax2.scatter(locs[up_inflx], values[up_inflx], color="g")
                ax2.scatter(locs[dwn_inflx], values[dwn_inflx], color="m")

            if len(peaks) == 0 or len(lows) < 2:
                continue

            if self.auto_picker.value:

                # Check dipping direction
                x_ind = np.min([values.shape[0] - 2, np.searchsorted(locs, center_x)])
                if values[x_ind] > values[x_ind + 1]:
                    cox = peaks[np.searchsorted(locs[peaks], center_x) - 1]
                else:
                    x_ind = np.min(
                        [peaks.shape[0] - 1, np.searchsorted(locs[peaks], center_x)]
                    )
                    cox = peaks[x_ind]

                start = lows[np.searchsorted(locs[lows], locs[cox]) - 1]
                end = lows[
                    np.min([np.searchsorted(locs[lows], locs[cox]), len(lows) - 1])
                ]

                bump_x = locs[start:end]
                bump_v = values[start:end]

                if len(bump_x) == 0:
                    continue

                inflx_up = np.searchsorted(
                    bump_x,
                    locs[up_inflx[np.searchsorted(locs[up_inflx], locs[cox]) - 1]],
                )
                inflx_up = np.max([0, inflx_up])

                inflx_dwn = np.searchsorted(
                    bump_x,
                    locs[dwn_inflx[np.searchsorted(locs[dwn_inflx], locs[cox])]],
                )
                inflx_dwn = np.min([bump_x.shape[0] - 1, inflx_dwn])

                peak = np.min([bump_x.shape[0] - 1, np.searchsorted(bump_x, locs[cox])])

                for ii, group in enumerate(self.group_list.value):
                    if channel in self.groups[group]["channels"]:
                        self.groups[group]["inflx_up"].append(
                            np.r_[bump_x[inflx_up], bump_v[inflx_up]]
                        )
                        self.groups[group]["peaks"].append(
                            np.r_[bump_x[peak], bump_v[peak]]
                        )
                        self.groups[group]["times"].append(times[channel])
                        self.groups[group]["inflx_dwn"].append(
                            np.r_[bump_x[inflx_dwn], bump_v[inflx_dwn]]
                        )
                        self.groups[group]["locations"].append(bump_x)
                        self.groups[group]["values"].append(bump_v)

                        # Compute average dip
                        left_ratio = (bump_v[peak] - bump_v[inflx_up]) / (
                            bump_x[peak] - bump_x[inflx_up]
                        )
                        right_ratio = (bump_v[peak] - bump_v[inflx_dwn]) / (
                            bump_x[inflx_dwn] - bump_x[peak]
                        )

                        if left_ratio > right_ratio:
                            ratio = right_ratio / left_ratio
                            ori = "left"
                        else:
                            ratio = left_ratio / right_ratio
                            ori = "right"

                        dip = np.rad2deg(np.arcsin(ratio))

                        # Left
                        ax2.plot(
                            bump_x[:peak],
                            bump_v[:peak],
                            "--",
                            color=self.groups[group]["color"],
                        )
                        # Right
                        ax2.plot(
                            bump_x[peak:],
                            bump_v[peak:],
                            color=self.groups[group]["color"],
                        )
                        ax2.scatter(
                            self.groups[group]["peaks"][-1][0],
                            self.groups[group]["peaks"][-1][1],
                            s=100,
                            c=self.groups[group]["color"],
                            marker=self.marker[ori],
                        )
                        if ~np.isnan(dip):
                            ax2.text(
                                self.groups[group]["peaks"][-1][0],
                                self.groups[group]["peaks"][-1][1],
                                f"{dip:.0f}",
                                va="bottom",
                                ha="center",
                            )
                        ax2.scatter(
                            self.groups[group]["inflx_dwn"][-1][0],
                            self.groups[group]["inflx_dwn"][-1][1],
                            s=100,
                            c=self.groups[group]["color"],
                            marker="1",
                        )
                        ax2.scatter(
                            self.groups[group]["inflx_up"][-1][0],
                            self.groups[group]["inflx_up"][-1][1],
                            s=100,
                            c=self.groups[group]["color"],
                            marker="2",
                        )

        ax2.plot(
            [
                self.lines.profile.locations_resampled[0],
                self.lines.profile.locations_resampled[-1],
            ],
            [0, 0],
            "r",
        )
        ax2.plot([center_x, center_x], [0, y_min], "r--")
        ax2.scatter(center_x, y_min, s=20, c="r", marker="^")

        for group in self.groups.values():
            if group["peaks"]:
                peaks = np.vstack(group["peaks"])
                ratio = peaks[:, 1] / peaks[0, 1]
                ind = np.where(ratio >= (threshold / 100))[0][-1]
                #                 print(ind)
                ax2.plot(
                    peaks[: ind + 1, 0], peaks[: ind + 1, 1], "--", color=group["color"]
                )
        #                 ax2.plot([peaks[0, 0], peaks[0, 0]], [peaks[0, 1], peaks[-1, 1]], '--', color='k')
        #                 ax2.plot(
        #                     [group['inflx_up'][ind][0], group['inflx_dwn'][ind][0]]
        #                     [group['inflx_up'][ind][1], group['inflx_dwn'][ind][1]], '--', color=[0.5,0.5,0.5]
        #                 )

        if scale == "symlog":
            plt.yscale("symlog", linthreshy=scale_value)

        x_lims = [
            center_x - focus / 2.0 * self.lines.profile.locations_resampled[-1],
            center_x + focus / 2.0 * self.lines.profile.locations_resampled[-1],
        ]
        ax2.set_xlim(x_lims)
        ax2.set_title(f"Line: {ind}")
        ax2.set_ylabel("dBdT")

        if x_label == "Easting":

            ax2.text(
                center_x,
                0,
                f"{self.lines.profile.interp_x(center_x):.0f} m E",
                va="top",
                ha="center",
                bbox={"edgecolor": "r"},
            )
            xlbl = [
                f"{self.lines.profile.interp_x(label):.0f}"
                for label in ax2.get_xticks()
            ]
            ax2.set_xticklabels(xlbl)
            ax2.set_xlabel("Easting (m)")
        elif x_label == "Northing":
            ax2.text(
                center_x,
                0,
                f"{self.lines.profile.interp_y(center_x):.0f} m N",
                va="top",
                ha="center",
                bbox={"edgecolor": "r"},
            )
            xlbl = [
                f"{self.lines.profile.interp_y(label):.0f}"
                for label in ax2.get_xticks()
            ]
            ax2.set_xticklabels(xlbl)
            ax2.set_xlabel("Northing (m)")
        else:
            ax2.text(
                center_x,
                0,
                f"{center_x:.0f} m",
                va="top",
                ha="center",
                bbox={"edgecolor": "r"},
            )
            ax2.set_xlabel("Distance (m)")

        ax2.grid(True)

        pos2 = ax2.get_position()

        ax = [pos2.x0, pos2.y0, pos2.width, pos2.height].copy()

        if self.auto_picker.value:
            ax[0] += 0.25
            ax[1] -= 0.5
            ax[2] /= 3
            ax[3] /= 2
            ax4 = plt.axes(ax)
            for group in self.group_list.value:
                if len(self.groups[group]["peaks"]) == 0:
                    continue

                peaks = (
                    np.vstack(self.groups[group]["peaks"])
                    * self.em_system_specs[self.system.value]["normalization"]
                )

                tc = np.hstack(self.groups[group]["times"][: ind + 1])
                vals = np.log(peaks[: ind + 1, 1])

                if tc.shape[0] < 2:
                    continue
                # Compute linear trend
                A = np.c_[np.ones_like(tc), tc]
                a, c = np.linalg.solve(np.dot(A.T, A), np.dot(A.T, vals))
                d = np.r_[tc.min(), tc.max()]
                vv = d * c + a

                ratio = np.abs((vv[0] - vv[1]) / (d[0] - d[1]))
                #                 angl = np.arctan(ratio**-1.)

                self.groups[group]["mad_tau"] = ratio ** -1.0

                ax4.plot(
                    d,
                    np.exp(d * c + a),
                    "--",
                    linewidth=2,
                    color=self.groups[group]["color"],
                )
                ax4.text(
                    np.mean(d),
                    np.exp(np.mean(vv)),
                    f"{ratio ** -1.:.2e}",
                    color=self.groups[group]["color"],
                )
                #                 plt.yscale('symlog', linthreshy=scale_value)
                #                 ax4.set_aspect('equal')
                ax4.scatter(
                    np.hstack(self.groups[group]["times"]),
                    peaks[:, 1],
                    color=self.groups[group]["color"],
                    marker="^",
                )
                ax4.grid(True)

            plt.yscale("symlog")
            ax4.yaxis.set_label_position("right")
            ax4.yaxis.tick_right()
            ax4.set_ylabel("log(V)")
            ax4.set_xlabel("Time (sec)")
            ax4.set_title("Decay - MADTau")

    def plot_model_selection(self, ind, center, focus):

        fig = plt.figure(figsize=(12, 8))
        ax3 = plt.subplot()

        if (
            getattr(self, "survey", None) is None
            or getattr(self.lines, "profile", None) is None
        ):
            return

        center_x = center * self.lines.profile.locations_resampled[-1]

        x_lims = [
            center_x - focus / 2.0 * self.lines.profile.locations_resampled[-1],
            center_x + focus / 2.0 * self.lines.profile.locations_resampled[-1],
        ]

        if getattr(self.lines, "model_x", None) is not None:
            return

            cs = ax3.tricontourf(
                self.lines.model_x,
                self.lines.model_z,
                self.lines.model_cells.reshape((-1, 3)),
                np.log10(self.lines.model_values),
                levels=np.linspace(-3, 0.5, 25),
                vmin=-2,
                vmax=-0.75,
                cmap="rainbow",
            )
            ax3.tricontour(
                self.lines.model_x,
                self.lines.model_z,
                self.lines.model_cells.reshape((-1, 3)),
                np.log10(self.lines.model_values),
                levels=np.linspace(-3, 0.5, 25),
                colors="k",
                linestyles="solid",
                linewidths=0.5,
            )
            #         ax3.scatter(center_x, center_z, 100, c='r', marker='x')
            ax3.set_xlim(x_lims)
            ax3.set_aspect("equal")
            ax3.grid(True)

    def line_update(self):
        """
        Re-compute derivatives
        """

        if getattr(self, "survey", None) is None:
            return

        if (
            len(self.survey.get_data(self.lines.value.value)) == 0
            or self.lines.lines.value == ""
        ):
            return

        line_ind = np.where(
            np.asarray(self.survey.get_data(self.lines.value.value)[0].values)
            == self.lines.lines.value
        )[0]

        if len(line_ind) == 0:
            return

        self.survey.line_indices = line_ind
        xyz = self.survey.vertices[line_ind, :]

        if np.std(xyz[:, 1]) > np.std(xyz[:, 0]):
            start = np.argmin(xyz[:, 1])
        else:
            start = np.argmin(xyz[:, 0])

        self.lines.profile = signal_processing_1d(
            xyz, None, smoothing=self.smoothing.value, residual=self.residual.value
        )

        # Get the corresponding along line model
        origin = xyz[0, :2]

        if self.workspace.get_entity(self.model_objects.value):
            surf_model = self.workspace.get_entity(self.model_objects.value)[0]

        if surf_model.get_data("Line") and np.any(
            np.where(surf_model.get_data("Line")[0].values == self.lines.lines.value)[0]
        ):

            surf_id = surf_model.get_data("Line")[0].values
            #             surf_ind = np.where(
            #                 surf_id == self.lines.lines.value
            #             )[0]

            cell_ind = np.where(
                surf_id[surf_model.cells[:, 0]] == self.lines.lines.value
            )[0]

            cells = surf_model.cells[cell_ind, :]
            vert_ind, cell_ind = np.unique(cells, return_inverse=True)

            surf_verts = surf_model.vertices[vert_ind, :]
            self.lines.model_x = np.linalg.norm(
                np.c_[
                    xyz[start, 0] - surf_verts[:, 0], xyz[start, 1] - surf_verts[:, 1]
                ],
                axis=1,
            )
            self.lines.model_z = surf_model.vertices[vert_ind, 2]
            self.lines.model_cells = cell_ind
            self.lines.model_values = surf_model.get_data(self.model_field.value)[
                0
            ].values[vert_ind]
        else:
            self.lines.model_x = None

    def reset_default_bounds(self):

        try:
            first, last = np.asarray(
                self.group_default_early.value.split("-"), dtype="int"
            )
            self.early = np.arange(first, last + 1).tolist()
        except ValueError:
            return

        try:
            first, last = np.asarray(
                self.group_default_middle.value.split("-"), dtype="int"
            )
            self.middle = np.arange(first, last + 1).tolist()
        except ValueError:
            return

        try:
            first, last = np.asarray(
                self.group_default_late.value.split("-"), dtype="int"
            )
            self.last = np.arange(first, last + 1).tolist()
        except ValueError:
            return

        for group in self.groups.values():
            gates = []
            if len(group["defaults"]) > 0:
                for default in group["defaults"]:
                    gates += getattr(self, default)
                group["gates"] = gates

        self.set_default_groups(self.channels.options)

    def show_model_trigger(self):
        """
        Add the model widget
        """
        if self.show_model.value:
            self._widget.children = [self.data_panel, self.show_model, self.model_panel]
        else:
            self._widget.children = [self.data_panel, self.show_model]

    @property
    def widget(self):
        return self._widget
